import abc
import pathlib
import shlex
from typing import Iterable, List, Optional, Set, Tuple

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.code import compile_item, run_item
from rbx.box.schema import CodeItem, GeneratorCall, Testcase, TestcaseSubgroup
from rbx.box.testcase_utils import (
    TestcaseEntry,
    TestcasePattern,
    fill_output_for_defined_testcase,
)
from rbx.grading.steps import DigestHolder, DigestOrDest, DigestOrSource


def _get_group_input(
    group_path: pathlib.Path, subgroup_prefix: str, i: int
) -> pathlib.Path:
    return group_path / f'{subgroup_prefix}{i:03d}.in'


def _get_group_output(
    group_path: pathlib.Path, subgroup_prefix: str, i: int
) -> pathlib.Path:
    return group_path / f'{subgroup_prefix}{i:03d}.out'


async def run_generator_script(testcase: TestcaseSubgroup) -> str:
    assert testcase.generatorScript is not None

    cacher = package.get_file_cacher()

    if not testcase.generatorScript.path.is_file():
        console.console.print(
            f'[error]Generator script not found: [item]{testcase.generatorScript.path}[/item][/error]'
        )
        raise typer.Exit(1)

    script_digest = DigestHolder()
    if testcase.generatorScript.path.suffix == '.txt':
        script_digest.value = cacher.put_file_from_path(testcase.generatorScript.path)
    else:
        try:
            compiled_digest = compile_item(testcase.generatorScript)
        except:
            console.console.print(
                f'[error]Failed compiling generator script for group [item]{testcase.name}[/item].[/error]'
            )
            raise

        run_stderr = DigestHolder()
        run_log = await run_item(
            testcase.generatorScript,
            DigestOrSource.create(compiled_digest),
            stdout=DigestOrDest.create(script_digest),
            stderr=DigestOrDest.create(run_stderr),
        )

        if run_log is None or run_log.exitcode != 0:
            console.console.print(
                f'Could not run generator script for group {testcase.name}'
            )
            if run_log is not None:
                console.console.print(
                    f'[error]Summary:[/error] {run_log.get_summary()}'
                )
            if run_stderr.value is not None:
                console.console.print('[error]Stderr:[/error]')
                console.console.print(
                    package.get_digest_as_string(run_stderr.value) or ''
                )
            raise typer.Exit(1)

    assert script_digest.value
    script = cacher.get_file_content(script_digest.value).decode()
    return script


def _extract_script_lines(script: str) -> Iterable[Tuple[str, str, int]]:
    lines = script.splitlines()
    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            continue
        if line.startswith('#'):
            continue
        yield shlex.split(line)[0], shlex.join(shlex.split(line)[1:]), i + 1


class GeneratorScriptEntry(BaseModel):
    path: pathlib.Path
    line: int

    def __str__(self) -> str:
        return f'{self.path}:{self.line}'


class GenerationMetadata(BaseModel):
    copied_to: Testcase

    copied_from: Optional[Testcase] = None
    generator_call: Optional[GeneratorCall] = None
    generator_script: Optional[GeneratorScriptEntry] = None


class GenerationTestcaseEntry(BaseModel):
    group_entry: TestcaseEntry
    subgroup_entry: TestcaseEntry

    metadata: GenerationMetadata
    validator: Optional[CodeItem] = None
    extra_validators: List[CodeItem] = []
    model_solution: Optional[CodeItem] = None


class TestcaseVisitor(abc.ABC):
    @abc.abstractmethod
    async def visit(self, entry: GenerationTestcaseEntry):
        pass

    def should_visit_group(self, group_name: str) -> bool:
        return True

    def should_visit_subgroup(self, subgroup_path: str) -> bool:
        return True

    def should_visit_generator_scripts(
        self, group_name: str, subgroup_path: str
    ) -> bool:
        return True


class TestcaseGroupVisitor(TestcaseVisitor):
    def __init__(self, groups: Optional[Set[str]] = None):
        self.groups = groups

    def should_visit_group(self, group_name: str) -> bool:
        return self.groups is None or group_name in self.groups


async def run_testcase_visitor(visitor: TestcaseVisitor):
    pkg = package.find_problem_package_or_die()

    async def _explore_subgroup(
        subgroup: TestcaseSubgroup,
        subgroup_index: Optional[int],
        prefix: List[str],
        validator: Optional[CodeItem] = None,
        extra_validators: Optional[List[CodeItem]] = None,
        model_solution: Optional[CodeItem] = None,
    ):
        extra_validators = extra_validators or []

        assert prefix and len(prefix) >= 1 and len(prefix) <= 2
        group_path = prefix[0]
        subgroup_path = '/'.join(prefix)
        if not visitor.should_visit_subgroup(subgroup_path):
            return

        def _entry(i: int) -> TestcaseEntry:
            return TestcaseEntry(group=group_path, index=i)

        def _sub_entry(i: int) -> TestcaseEntry:
            return TestcaseEntry(group=subgroup_path, index=i)

        def _copied_to(i: int) -> Testcase:
            group_fs_path = package.get_build_testgroup_path(group_path)
            group_prefix = ''
            if subgroup_index is not None:
                group_prefix = f'{subgroup_index}-'
            if len(prefix) == 2:
                group_prefix += f'{prefix[1]}-'
            return Testcase(
                inputPath=_get_group_input(group_fs_path, group_prefix, i),
                outputPath=_get_group_output(group_fs_path, group_prefix, i),
            )

        # Go through testcases.
        i = 0
        # Individual testcases.
        for tc in subgroup.testcases or []:
            await visitor.visit(
                GenerationTestcaseEntry(
                    group_entry=_entry(i),
                    subgroup_entry=_sub_entry(i),
                    metadata=GenerationMetadata(
                        copied_from=fill_output_for_defined_testcase(tc),
                        copied_to=_copied_to(i),
                    ),
                    validator=validator,
                    extra_validators=extra_validators,
                    model_solution=model_solution,
                )
            )
            i += 1

        # Glob testcases.
        if subgroup.testcaseGlob:
            matched_inputs = sorted(pathlib.PosixPath().glob(subgroup.testcaseGlob))

            for input_path in matched_inputs:
                if not input_path.is_file() or input_path.suffix != '.in':
                    continue

                tc = Testcase(inputPath=input_path)
                await visitor.visit(
                    GenerationTestcaseEntry(
                        group_entry=_entry(i),
                        subgroup_entry=_sub_entry(i),
                        metadata=GenerationMetadata(
                            copied_from=fill_output_for_defined_testcase(tc),
                            copied_to=_copied_to(i),
                        ),
                        validator=validator,
                        extra_validators=extra_validators,
                        model_solution=model_solution,
                    )
                )
                i += 1

        # Single generators.
        for generator_call in subgroup.generators:
            await visitor.visit(
                GenerationTestcaseEntry(
                    group_entry=_entry(i),
                    subgroup_entry=_sub_entry(i),
                    metadata=GenerationMetadata(
                        generator_call=generator_call,
                        copied_to=_copied_to(i),
                    ),
                    validator=validator,
                    extra_validators=extra_validators,
                    model_solution=model_solution,
                )
            )
            i += 1

        if not visitor.should_visit_generator_scripts(group_path, subgroup_path):
            return

        # Run generator script.
        if subgroup.generatorScript is not None:
            script = await run_generator_script(subgroup)

            # Run each line from generator script.
            for generator_name, args, line_number in _extract_script_lines(script):
                call = GeneratorCall(name=generator_name, args=args)
                await visitor.visit(
                    GenerationTestcaseEntry(
                        group_entry=_entry(i),
                        subgroup_entry=_sub_entry(i),
                        metadata=GenerationMetadata(
                            generator_call=call,
                            generator_script=GeneratorScriptEntry(
                                path=subgroup.generatorScript.path,
                                line=line_number,
                            ),
                            copied_to=_copied_to(i),
                        ),
                        validator=validator,
                        extra_validators=extra_validators,
                        model_solution=model_solution,
                    )
                )
                i += 1

    for group in pkg.testcases:
        if not visitor.should_visit_group(group.name):
            continue

        group_validator = pkg.validator
        if group.validator is not None:
            group_validator = group.validator

        extra_validators = group.extraValidators
        await _explore_subgroup(
            group,
            0 if group.subgroups else None,
            [group.name],
            validator=group_validator,
            extra_validators=extra_validators,
            model_solution=group.model_solution,
        )

        for i, subgroup in enumerate(group.subgroups):
            await _explore_subgroup(
                subgroup,
                i + 1,
                [group.name, subgroup.name],
                validator=group_validator,
                extra_validators=extra_validators + subgroup.extraValidators,
                model_solution=group.model_solution,
            )


async def extract_generation_testcases(
    entries: List[TestcaseEntry],
) -> List[GenerationTestcaseEntry]:
    # TODO: support subgroups.
    groups = set(entry.group for entry in entries)
    entry_keys = set(entry.key() for entry in entries)

    res: List[GenerationTestcaseEntry] = []

    class ExtractGenerationTestcasesVisitor(TestcaseVisitor):
        def should_visit_group(self, group_name: str) -> bool:
            return group_name in groups

        async def visit(self, entry: GenerationTestcaseEntry):
            # TODO: support subgroups.
            if entry.group_entry.key() not in entry_keys:
                return
            res.append(entry)

    await run_testcase_visitor(ExtractGenerationTestcasesVisitor())
    return res


async def extract_generation_testcases_from_groups(
    groups: Optional[Set[str]] = None,
) -> List[GenerationTestcaseEntry]:
    res: List[GenerationTestcaseEntry] = []

    class ExtractGenerationTestcasesVisitor(TestcaseGroupVisitor):
        async def visit(self, entry: GenerationTestcaseEntry):
            res.append(entry)

    await run_testcase_visitor(ExtractGenerationTestcasesVisitor(groups))
    return res


async def extract_generation_testcases_from_patterns(
    patterns: List[TestcasePattern],
) -> List[GenerationTestcaseEntry]:
    res: List[GenerationTestcaseEntry] = []

    class ExtractGenerationTestcasesVisitor(TestcaseVisitor):
        def should_visit_group(self, group_name: str) -> bool:
            return any(pattern.intersecting_group(group_name) for pattern in patterns)

        def should_visit_subgroup(self, subgroup_path: str) -> bool:
            return any(
                pattern.intersecting_group(subgroup_path) for pattern in patterns
            )

        async def visit(self, entry: GenerationTestcaseEntry):
            if not any(
                pattern.match(entry.group_entry) for pattern in patterns
            ) and not any(pattern.match(entry.subgroup_entry) for pattern in patterns):
                return
            res.append(entry)

    await run_testcase_visitor(ExtractGenerationTestcasesVisitor())
    return res



import os
import pathlib
import tempfile
from typing import Any, Dict, Optional

import rich
import rich.progress
import typer

from rbx import console
from rbx.box import header, limits_info, package
from rbx.box.generators import get_all_built_testcases
from rbx.box.lang import code_to_langs, is_valid_lang_code
from rbx.box.packaging.polygon import polygon_api as api
from rbx.box.schema import CodeItem, ExpectedOutcome, Solution, TaskType, Testcase
from rbx.box.statements.build_statements import get_relative_assets
from rbx.box.statements.builders import (
    StatementBlocks,
    StatementBuilderProblem,
    render_jinja_blocks,
)
from rbx.box.statements.schema import Statement, StatementType
from rbx.box.testcase_utils import (
    TestcaseInteractionParsingError,
    get_alternate_interaction_texts,
    parse_interaction,
)

_API_URL = 'https://polygon.codeforces.com/api'

POLY = api.Polygon(
    _API_URL,
    os.environ.get('POLYGON_API_KEY', '').strip(),
    os.environ.get('POLYGON_API_SECRET', '').strip(),
)


def _get_source_type(code: CodeItem):
    return None


def _get_solution_tag(solution: Solution, is_first: bool = False) -> api.SolutionTag:
    if solution.outcome == ExpectedOutcome.ACCEPTED:
        return api.SolutionTag.OK if not is_first else api.SolutionTag.MA
    if solution.outcome == ExpectedOutcome.ACCEPTED_OR_TLE:
        return api.SolutionTag.TO
    if solution.outcome == ExpectedOutcome.WRONG_ANSWER:
        return api.SolutionTag.WA
    if solution.outcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED:
        return api.SolutionTag.TL
    if solution.outcome == ExpectedOutcome.MEMORY_LIMIT_EXCEEDED:
        return api.SolutionTag.ML
    if solution.outcome == ExpectedOutcome.RUNTIME_ERROR:
        return api.SolutionTag.RE
    return api.SolutionTag.RJ


def _find_or_create_problem(problem_name: str) -> api.Problem:
    results = POLY.problems_list(name=problem_name)
    for result in results:
        if result.name == problem_name:
            console.console.print(
                f'Found already existing problem [item]{problem_name}[/item].'
            )
            return result
    console.console.print(f'Creating new problem [item]{problem_name}[/item].')
    return POLY.problem_create(problem_name)


def _update_problem_info(problem: api.Problem):
    pkg = package.find_problem_package_or_die()

    problem.update_info(
        api.ProblemInfo(
            interactive=pkg.type == TaskType.COMMUNICATION,
            time_limit=pkg.timeLimit,
            memory_limit=pkg.memoryLimit,
        )
    )


def _get_checker_name() -> str:
    checker = package.get_checker()
    return checker.path.with_stem('checker').name


def _get_interactor_name() -> str:
    interactor = package.get_interactor()
    return interactor.path.with_stem('interactor').name


def _get_validator_name() -> str:
    validator = package.get_validator()
    return validator.path.with_stem('validator').name


def _update_rbx_header(problem: api.Problem):
    console.console.print('Uploading rbx.h...')
    rbx_header = header.get_header()
    problem.save_file(
        type=api.FileType.RESOURCE,
        name='rbx.h',
        file=rbx_header.read_bytes(),
        source_type=None,
    )


def _update_checker(problem: api.Problem):
    console.console.print('Uploading checker...')
    checker = package.get_checker()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_checker_name(),
        file=checker.path.read_bytes(),
        source_type=_get_source_type(checker),
    )

    problem.set_checker(_get_checker_name())


def _update_interactor(problem: api.Problem):
    console.console.print('Uploading interactor...')
    interactor = package.get_interactor()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_interactor_name(),
        file=interactor.path.read_bytes(),
        source_type=_get_source_type(interactor),
    )

    problem.set_interactor(_get_interactor_name())


def _upload_validator(problem: api.Problem):
    console.console.print('Uploading validator...')
    validator = package.get_validator()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_validator_name(),
        file=validator.path.read_bytes(),
        source_type=_get_source_type(validator),
    )

    problem.set_validator(_get_validator_name())


def _save_skip_coinciding_testcases(problem: api.Problem, *args, **kwargs) -> bool:
    try:
        problem.save_test(*args, **kwargs)
    except api.PolygonRequestFailedException as e:
        if 'test coincides with' in e.comment.lower():
            return False
        raise
    return True


def _get_test_params_for_statement(
    testcase: Testcase, is_sample: bool
) -> Dict[str, Any]:
    if not is_sample:
        return {}
    res: Dict[str, Any] = {'test_use_in_statements': True}
    if testcase.outputPath is not None:
        res['test_output_for_statements'] = testcase.outputPath.read_text()
    else:
        return res

    pio_path = testcase.outputPath.with_suffix('.pio')
    if pio_path.is_file():
        try:
            interaction = parse_interaction(pio_path)
        except TestcaseInteractionParsingError:
            pass
        else:
            res['test_input_for_statements'], res['test_output_for_statements'] = (
                get_alternate_interaction_texts(interaction)
            )
            return res

    # .pio does not exist or is not parseable, fallback to .pin and .pout.
    pin_path = testcase.outputPath.with_suffix('.pin')
    if pin_path.is_file():
        res['test_input_for_statements'] = pin_path.read_text()
    pout_path = testcase.outputPath.with_suffix('.pout')
    if pout_path.is_file():
        res['test_output_for_statements'] = pout_path.read_text()
    return res


def _upload_testcases(problem: api.Problem):
    pkg = package.find_problem_package_or_die()
    testcases = get_all_built_testcases()
    i = 0

    with rich.progress.Progress(speed_estimate_period=5) as progress:
        total_len = 0
        for group in pkg.testcases:
            total_len += len(testcases[group.name])
        task_id = progress.add_task('Uploading testcases...', total=total_len)
        for group in pkg.testcases:
            for testcase in testcases[group.name]:
                is_sample = group.name == 'samples'
                saved = _save_skip_coinciding_testcases(
                    problem,
                    testset='tests',
                    test_index=i + 1,
                    test_input=testcase.inputPath.read_text(),
                    **_get_test_params_for_statement(testcase, is_sample),
                )
                progress.update(task_id, advance=1)
                if saved:
                    i += 1


def _upload_solutions(problem: api.Problem):
    console.console.print('Uploading main solution...')
    main_solution = package.get_main_solution()
    if main_solution is None or main_solution.outcome != ExpectedOutcome.ACCEPTED:
        return
    problem.save_solution(
        main_solution.path.name,
        main_solution.path.read_bytes(),
        source_type=_get_source_type(main_solution),
        tag=api.SolutionTag.MA,
    )

    for i, solution in enumerate(package.get_solutions()):
        console.console.print(
            f'Uploading solution [item]{solution.path.name}[/item] (tag: [item]{_get_solution_tag(solution, is_first=i == 0)}[/item])...'
        )
        problem.save_solution(
            solution.path.name,
            solution.path.read_bytes(),
            source_type=_get_source_type(solution),
            tag=_get_solution_tag(solution, is_first=i == 0),
        )


def _get_statement_for_language(language: str) -> Optional[Statement]:
    pkg = package.find_problem_package_or_die()
    for statement in pkg.expanded_statements:
        if statement.language == language:
            return statement
    return None


def _get_statement_blocks(statement: Statement) -> StatementBlocks:
    # TODO: actually try to convert to rbxTeX
    assert statement.type == StatementType.rbxTeX
    pkg = package.find_problem_package_or_die()
    # TODO: pull this from a library, too hacky at the moment
    builder_problem = StatementBuilderProblem(
        limits=limits_info.get_limits_profile(profile='polygon'),
        package=pkg,
        statement=statement,
        vars={
            **pkg.expanded_vars,
            **statement.expanded_vars,
        },
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        return render_jinja_blocks(
            pathlib.Path(temp_dir),
            statement.path.read_bytes(),
            **builder_problem.build_inner_jinja_kwargs(),
        )


def _get_explanations(explanations: Dict[int, str]) -> str:
    entries = []
    for i in sorted(explanations):
        explanation = explanations[i]
        entries.append(f'\\textbf{{Explanation for example {i + 1}}}\n\n{explanation}')
    return '\n\n'.join(entries)


def _get_notes_with_explanations(blocks: StatementBlocks) -> Optional[str]:
    notes = blocks.blocks.get('notes')
    explanations = blocks.explanations
    if notes is None and not explanations:
        return None
    if notes is None:
        return _get_explanations(blocks.explanations)
    return notes + '\n\n' + _get_explanations(blocks.explanations)


def _upload_statement_resources(problem: api.Problem, statement: Statement):
    assets = get_relative_assets(statement.path, statement.assets)
    for asset, relative_asset in assets:
        console.console.print(
            f'Uploading statement resource [item]{relative_asset}[/item]...'
        )
        resource_bytes = asset.read_bytes()
        if len(resource_bytes) >= 1024 * 1024:  # >= 1mb
            console.console.print(
                f'[error]Statement resource [item]{relative_asset}[/item] is too large to upload (more than 1MB).[/error]'
            )
            raise typer.Exit(1)
        problem.save_statement_resource(
            name=str(relative_asset),
            file=resource_bytes,
        )


def _upload_statement(
    problem: api.Problem, main_language: Optional[str], upload_as_english: bool = False
):
    pkg = package.find_problem_package_or_die()

    lang_list = []
    languages = set()
    for statement in pkg.expanded_statements:
        if not is_valid_lang_code(statement.language):
            continue
        languages.add(statement.language)
        lang_list.append(statement.language)
    uploaded_languages = set()

    if main_language is None:
        main_language = lang_list[0]

    # Put the main language first.
    lang_list = list(languages)
    for i in range(len(lang_list)):
        if lang_list[i] == main_language:
            lang_list[i], lang_list[0] = lang_list[0], lang_list[i]
            break

    # Prioritize English statements.
    for language in lang_list:
        statement = _get_statement_for_language(language)
        if statement is None:
            continue
        if statement.type != StatementType.rbxTeX:
            continue
        statement_lang = code_to_langs([language])[0]
        console.console.print(
            f'Uploading statement for language [item]{language}[/item] (polygon language: [item]{statement_lang}[/item])...'
        )
        uploaded_language = statement_lang
        if main_language == language:
            if not upload_as_english:
                console.console.print(
                    '[warning]By default, Polygon statements are uploaded respecting their original language.\n'
                    'Codeforces does not work well with statements in other languages. If you want a better experience, '
                    'use the [item]--upload-as-english[/item] option to force the main statement to be uploaded in English.[/warning]'
                )
            else:
                uploaded_language = 'english'
        if uploaded_language in uploaded_languages:
            continue
        uploaded_languages.add(uploaded_language)
        blocks = _get_statement_blocks(statement)
        polygon_statement = api.Statement(
            encoding='utf-8',
            name=statement.title,
            legend=blocks.blocks.get('legend') or '',
            input=blocks.blocks.get('input') or '',
            output=blocks.blocks.get('output') or '',
            interaction=(blocks.blocks.get('interaction') or '')
            if pkg.type == TaskType.COMMUNICATION
            else None,
            notes=_get_notes_with_explanations(blocks) or '',
        )
        problem.save_statement(
            lang=uploaded_language,
            problem_statement=polygon_statement,
        )

        _upload_statement_resources(problem, statement)


def _normalize_problem_name(name: str) -> str:
    return name.replace(' ', '-').replace('_', '-').lower()


async def upload_problem(
    name: str, main_language: Optional[str], upload_as_english: bool = False
):
    pkg = package.find_problem_package_or_die()
    name = _normalize_problem_name(name)
    problem = _find_or_create_problem(name)
    _update_problem_info(problem)
    _update_rbx_header(problem)
    _update_checker(problem)

    if (
        pkg.type == TaskType.COMMUNICATION
        and package.get_interactor_or_nil() is not None
    ):
        _update_interactor(problem)

    # if pkg.validator is not None:
    #     _upload_validator(problem)

    _upload_solutions(problem)
    _upload_testcases(problem)
    _upload_statement(
        problem, main_language=main_language, upload_as_english=upload_as_english
    )

    # Commit.
    console.console.print('Committing changes...')
    problem.commit_changes()

    console.console.print(
        f'[success]Problem [item]{name}[/item] uploaded successfully![/success]'
    )



import hashlib
import pathlib
from typing import IO


class Digester:
    """Simple wrapper of hashlib using our preferred hasher."""

    def __init__(self):
        self._hasher = hashlib.sha1()

    def update(self, b):
        """Add the bytes b to the hasher."""
        self._hasher.update(b)

    def digest(self):
        """Return the digest as an hex string."""
        return self._hasher.digest().hex()


def digest_cooperatively_into_digester(
    f: IO[bytes], digester: Digester, chunk_size: int = 2**20
):
    buf = f.read(chunk_size)
    while len(buf) > 0:
        digester.update(buf)
        buf = f.read(chunk_size)


def digest_cooperatively(f: IO[bytes], chunk_size: int = 2**20):
    d = Digester()
    digest_cooperatively_into_digester(f, d, chunk_size)
    return d.digest()


def digest_file(path: pathlib.Path):
    with open(path, 'rb') as f:
        return digest_cooperatively(f)



import pathlib
from typing import Annotated, Optional

import typer

from rbx import console, utils
from rbx.box import package, presets


def create(
    name: Annotated[
        str,
        typer.Argument(
            help='The name of the problem package to create. This will also be the name of the folder.'
        ),
    ],
    preset: Annotated[
        Optional[str],
        typer.Option(
            '--preset',
            '-p',
            help='Which preset to use to create this package. Can be a named of an already installed preset, or an URI, in which case the preset will be downloaded.',
        ),
    ] = None,
    path: Optional[pathlib.Path] = None,
):
    console.console.print(f'Creating new problem [item]{name}[/item]...')

    fetch_info = presets.get_preset_fetch_info_with_fallback(preset)
    dest_path = path or pathlib.Path(name)

    if dest_path.exists():
        console.console.print(
            f'[error]Directory [item]{dest_path}[/item] already exists.[/error]'
        )
        raise typer.Exit(1)

    presets.install_problem(dest_path, fetch_info)

    # Change problem name.
    ru, problem = package.get_ruyaml(dest_path)
    problem['name'] = name
    utils.save_ruyaml(dest_path / 'problem.rbx.yml', ru, problem)

    # fix_package(dest_path)

    presets.generate_lock(dest_path)



from typing import Optional, Tuple

from rbx.box import package
from rbx.box.contest import contest_package
from rbx.box.contest.schema import ContestProblem


def get_problem_entry_in_contest() -> Optional[Tuple[int, ContestProblem]]:
    contest = contest_package.find_contest_package()
    if contest is None:
        return None
    problem_path = package.find_problem()
    contest_path = contest_package.find_contest()

    for i, problem in enumerate(contest.problems):
        if problem.path is None:
            continue
        if (problem_path / 'problem.rbx.yml').samefile(
            contest_path / problem.path / 'problem.rbx.yml'
        ):
            return i, problem
    return None


def get_problem_shortname() -> Optional[str]:
    entry = get_problem_entry_in_contest()
    if entry is None:
        return None
    _, problem = entry
    return problem.short_name


def get_problem_index() -> Optional[int]:
    entry = get_problem_entry_in_contest()
    if entry is None:
        return None
    return entry[0]


def get_problem_name_with_contest_info() -> str:
    problem = package.find_problem_package_or_die()
    contest = contest_package.find_contest_package()
    short_name = get_problem_shortname()
    if contest is None or short_name is None:
        return problem.name
    return f'{contest.name}-{short_name}-{problem.name}'



import datetime
import functools
import hashlib
import os
import pathlib
import re
import shutil
import time
import typing
from typing import Any, Callable, List, NoReturn, Optional, Tuple, Union

import dateparser
import mechanize
import typer
from bs4 import BeautifulSoup, Tag
from pydantic import BaseModel
from throttlex import Throttler

from rbx import console
from rbx.box import naming
from rbx.box.tooling.boca.debug_utils import pretty_print_request_data
from rbx.grading.steps import Outcome

ALERT_REGEX = re.compile(r'\<script[^\>]*\>\s*alert\(\'([^\']+)\'\);?\s*\<\/script\>')
REDIRECT_REGEX = re.compile(
    r'\<script[^\>]*\>\s*document\.location\s*=\s*\'([^\']+)\'\;\s*\<\/script\>'
)
START_DATE_REGEX = re.compile(r'Start date\s*\(contest\=([^\)]+)\)')
UPLOAD_LOG_REGEX = re.compile(r'Problem (\d+) \([^\)]+\) updated')


def _parse_env_var(var: str, override: Optional[str]) -> str:
    if override is not None:
        return override
    value = os.environ.get(var)
    if value is None:
        console.console.print(
            f'[error][item]{var}[/item] is not set. Set it as an environment variable.[/error]'
        )
        raise typer.Exit(1)
    return value


def _parse_answer_as_outcome(answer: str) -> Optional[Outcome]:
    answer = answer.lower()
    if 'yes' in answer:
        return Outcome.ACCEPTED
    if 'wrong answer' in answer:
        return Outcome.WRONG_ANSWER
    if 'time limit exceeded' in answer:
        return Outcome.TIME_LIMIT_EXCEEDED
    if 'runtime error' in answer:
        return Outcome.RUNTIME_ERROR
    return None


def _big_hex_sum(hex1: str, hex2: str) -> str:
    return f'{int(hex1, 16) + int(hex2, 16):x}'


class BocaProblem(BaseModel):
    index: int
    shortname: str
    fullname: str
    basename: str
    color: str
    color_name: str
    descfile_url: str
    package_url: str
    package_hash: str


class BocaLanguage(BaseModel):
    index: int
    name: str
    extension: str


class BocaRun(BaseModel):
    run_number: int
    site_number: int
    problem_shortname: str
    outcome: Optional[Outcome]
    time: int
    status: str

    user: Optional[str] = None

    @classmethod
    def from_run_number(cls, run_number: int, site_number: int):
        return cls(
            run_number=run_number,
            site_number=site_number,
            problem_shortname='',
            outcome=None,
            time=0,
            status='',
        )


class BocaDetailedRun(BocaRun):
    language_repr: str
    code: str
    filename: pathlib.Path

    autojudge_answer: str


class BocaScraper:
    def __init__(
        self,
        base_url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        throttler: Optional[Throttler] = None,
        verbose: bool = False,
        is_judge: bool = False,
    ):
        self.base_url = _parse_env_var('BOCA_BASE_URL', base_url)
        self.username = _parse_env_var('BOCA_USERNAME', username)
        self.password = _parse_env_var('BOCA_PASSWORD', password)
        self.verbose = verbose
        self.loggedIn = False
        self.is_judge = is_judge
        self.throttler = throttler or Throttler(max_req=1, period=1)
        self.br = mechanize.Browser()
        self.br.set_handle_robots(False)
        self.br.addheaders = [  # type: ignore
            (
                'User-agent',
                'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.1) Gecko/2008071615 Fedora/3.0.1-1.fc9 Firefox/3.0.1',
            )
        ]

    def log(self, message: str):
        if self.verbose:
            console.console.print(message)

    def error(self, message: str) -> NoReturn:
        console.console.print(
            f'[error]{message} (at [item]{self.base_url}[/item])[/error]',
        )
        raise typer.Exit(1)

    def raw_error(self, message: str) -> NoReturn:
        console.console.print(f'[error]{message}[/error]')
        raise typer.Exit(1)

    def pretty_print(self, html: str):
        soup = BeautifulSoup(html, 'html.parser')
        console.console.print(soup.prettify())

    def get_redirect(self, html: str) -> Optional[str]:
        redirect = REDIRECT_REGEX.search(html)
        if redirect is None:
            return None
        return redirect.group(1)

    def get_alert(self, html: str) -> Optional[str]:
        alert = ALERT_REGEX.search(html)
        if alert is None:
            return None
        return alert.group(1)

    def log_response_alert(
        self,
        response: Any,
        message: str,
        alert_ok_fn: Optional[Callable[[str], bool]] = None,
    ) -> Tuple[Any, str]:
        if response is None:
            self.raw_error(
                f'{message} ([item]{self.base_url}[/item]):\nNo response received.'
            )
        html = response.read().decode()
        alert = self.get_alert(html)
        if alert:
            if alert_ok_fn is not None and alert_ok_fn(alert):
                return response, html
            self.raw_error(f'{message} ([item]{self.base_url}[/item]):\n{alert}')
        return response, html

    def check_logs_for_update(self, problem_id: int) -> bool:
        _, html = self.open(
            f'{self.base_url}/admin/log.php',
            error_msg='Error while checking whether package upload was successful',
        )

        soup = BeautifulSoup(html, 'html.parser')
        table = soup.select_one('table:nth-of-type(3)')
        if table is None:
            self.raw_error(
                'Error while checking whether package upload was successful:\nNo logs table found.'
            )
        rows = table.select('tr:not(:first-child)')
        for row in rows:
            # Check whether the log line is recent.
            date_cell = row.select('td:nth-of-type(5)')
            if date_cell is None:
                continue
            date = date_cell[0].text.strip()

            parsed_date = dateparser.parse(
                date, settings={'TO_TIMEZONE': 'UTC', 'RETURN_AS_TIMEZONE_AWARE': True}
            )
            if parsed_date is None:
                console.console.print(
                    f'Error while checking whether package upload was successful:\nCould not parse date [item]{date}[/item].'
                )
                raise typer.Exit(1)
            if parsed_date < datetime.datetime.now(
                datetime.timezone.utc
            ) - datetime.timedelta(minutes=1):
                continue

            # Check if the log line contains the problem id.
            log_cell = row.select('td:nth-of-type(6)')
            if log_cell is None:
                continue
            log_line = log_cell[0].text.strip()
            match = UPLOAD_LOG_REGEX.match(log_line)
            if match is None:
                continue
            found_id = int(match.group(1))
            if found_id == problem_id:
                return True
        return False

    def check_submit(self, response: Any, problem_id: int) -> bool:
        if response is None:
            self.raw_error(
                'Error while submitting problem to BOCA website:\nNo response received.'
            )
        html = response.read().decode()
        alert = self.get_alert(html)
        if alert:
            if 'Violation' in alert:
                return False
            self.raw_error(f'Error while submitting problem to BOCA website:\n{alert}')

        redirect = self.get_redirect(html)
        if redirect is None:
            self.raw_error(
                'Error while submitting problem to BOCA website:\nNo redirect found after upload.'
            )
        _, html = self.open(
            redirect, error_msg='Error while freeing BOCA problems after upload'
        )
        return self.check_logs_for_update(problem_id)

    def open(
        self,
        url_or_request: Union[str, mechanize.Request],
        *args,
        error_msg: Optional[str] = None,
        **kwargs,
    ):
        url_or_request = self.throttler.throttle(url_or_request)
        if isinstance(url_or_request, mechanize.Request):
            url = url_or_request.get_full_url()
        else:
            url = url_or_request
        if error_msg is None:
            error_msg = f'Error while opening [item]{url}[/item]'
        response = self.br.open(url_or_request, *args, **kwargs)
        return self.log_response_alert(response, error_msg)

    def hash_single(self, password: str) -> str:
        pwd_hash = hashlib.sha256(password.encode()).hexdigest()
        return pwd_hash

    def hash(self, password: str, salt: str) -> str:
        pwd_hash = self.hash_single(password)
        pwd_hash = self.hash_single(pwd_hash + salt)
        return pwd_hash

    def hash_two(self, password1: str, password2: str) -> str:
        return _big_hex_sum(self.hash_single(password1), self.hash_single(password2))

    def find_salt(self, html: str, field: str) -> str:
        needle = f"js_myhash(document.{field}.value)+'"
        start = html.index(needle)
        start_salt = start + len(needle)
        end_salt = html.index("'", start_salt)
        salt = html[start_salt:end_salt]
        return salt

    def login(self):
        if self.loggedIn:
            return

        _, html = self.open(
            f'{self.base_url}', error_msg='Error while opening BOCA login page'
        )

        salt = self.find_salt(html, 'form1.password')
        console.console.print(f'Using salt [item]{salt}[/item]')

        pwd_hash = self.hash(self.password, salt)

        login_url = f'{self.base_url}?name={self.username}&password={pwd_hash}'
        self.open(login_url, error_msg='Error while logging in to BOCA')

        self.loggedIn = True

    def upload(
        self,
        file: pathlib.Path,
        testing: bool = False,
    ) -> bool:
        self.open(
            f'{self.base_url}/admin/problem.php',
            error_msg='Error while opening BOCA problem upload page',
        )
        try:
            self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Problem upload form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)

        if testing:
            problem_index = 0
            problem_shortname = 'A'
            hex_color = None
        else:
            problem_index = naming.get_problem_index()
            if problem_index is None:
                console.console.print(
                    'It seems this problem is not part of a contest. Cannot upload it to BOCA.'
                )
                raise typer.Exit(1)

            problem_shortname = naming.get_problem_shortname()
            assert problem_shortname is not None
            if problem_shortname is None:
                console.console.print(
                    'It seems this problem is not part of a contest. Cannot upload it to BOCA.'
                )
                raise typer.Exit(1)

            entry = naming.get_problem_entry_in_contest()
            assert entry is not None
            _, problem_entry = entry

            hex_color = problem_entry.hex_color

        if hex_color is None:
            form['colorname'] = 'black'
            form['color'] = '000000'
        else:
            assert problem_entry.color_name is not None
            form['colorname'] = problem_entry.color_name.lower()
            form['color'] = hex_color[1:]

        form['problemnumber'] = f'{problem_index + 1}'
        form['problemname'] = problem_shortname
        form['confirmation'] = 'confirm'
        form['autojudge_new_sel'] = ['all']
        form['Submit3'] = 'Send'

        with file.open('rb') as f:
            form.add_file(
                f,
                filename=file.name,
                name='probleminput',
                content_type='application/zip',
            )
            response = self.br.submit()

        return self.check_submit(response, problem_index + 1)

    def login_and_upload(self, file: pathlib.Path):
        RETRIES = 3

        tries = 0
        ok = False
        while tries < RETRIES:
            tries += 1

            console.console.print('Logging in to BOCA...')
            self.login()
            console.console.print('Uploading problem to BOCA...')
            if not self.upload(file):
                console.console.print(
                    f'[warning]Potentially transient error while uploading problem to BOCA. Retrying ({tries}/{RETRIES})...[/warning]'
                )
                self.loggedIn = False
                continue

            ok = True
            console.console.print(
                '[success]Problem sent to BOCA. [item]rbx[/item] cannot determine the upload succeeded, check the website to be sure.[/success]'
            )
            break

        if not ok:
            console.console.print(
                '[error]Persistent error while uploading problem to BOCA website.[/error]'
            )
            console.console.print(
                '[warning]This might be caused by PHP max upload size limit (which usually defaults to 2MB).[/warning]'
            )
            console.console.print(
                '[warning]Check [item]https://www.php.net/manual/en/ini.core.php#ini.sect.file-uploads[/item] for more information.[/warning]'
            )
            raise typer.Exit(1)

    def get_link_from_a(self, a: Tag) -> Optional[str]:
        href = a.attrs.get('href')
        if href is None:
            return None
        link = self.br.find_link(url=href)
        if link is None:
            return None
        return link.absolute_url

    def get_first_link(self, tag: Tag) -> Optional[str]:
        a = tag.find('a')
        if a is None:
            return None
        return self.get_link_from_a(typing.cast(Tag, a))

    def list_problems(self) -> List[BocaProblem]:
        _, html = self.open(
            f'{self.base_url}/admin/problem.php',
            error_msg='Error while listing problems in BOCA',
        )

        soup = BeautifulSoup(html, 'html.parser')
        rows = soup.select('form[name="form0"] > table > tr')
        problems: List[BocaProblem] = []
        for row in rows[2:]:
            cells = row.select('& > td')
            index = int(cells[0].text.strip())
            shortname = cells[1].text.strip()
            fullname = cells[2].text.strip()
            basename = cells[3].text.strip()
            descfile_url = self.get_first_link(cells[4])
            package_url = self.get_first_link(cells[5])

            hash_balloon = cells[5].find('img')
            if descfile_url is None or package_url is None or hash_balloon is None:
                self.log(f'Skipping problem {shortname} because of missing data')
                continue
            hash_balloon = typing.cast(Tag, hash_balloon)
            hash = hash_balloon.attrs.get('alt')
            if hash is None:
                self.log(f'Skipping problem {shortname} because hash is None')
                continue
            hash = str(hash).strip()

            color_balloon = cells[6].find('img')
            if color_balloon is None:
                self.log(f'Skipping problem {shortname} because color balloon is None')
                continue
            color_balloon = typing.cast(Tag, color_balloon)
            color = color_balloon.attrs.get('alt')
            if color is None:
                self.log(f'Skipping problem {shortname} because color is None')
                continue
            color = str(color).strip()

            color_name = color_balloon.attrs.get('title')
            if color_name is None:
                self.log(f'Skipping problem {shortname} because color name is None')
                continue
            color_name = str(color_name).strip()

            problems.append(
                BocaProblem(
                    index=index,
                    shortname=shortname,
                    fullname=fullname,
                    basename=basename,
                    color=color,
                    color_name=color_name,
                    descfile_url=descfile_url,
                    package_url=package_url,
                    package_hash=hash,
                )
            )

        return problems

    def list_languages(self) -> List[BocaLanguage]:
        _, html = self.open(
            f'{self.base_url}/admin/language.php',
            error_msg='Error while listing languages in BOCA',
        )

        soup = BeautifulSoup(html, 'html.parser')
        tables = soup.select('table')
        for table in tables:
            if 'Language #' in table.text:
                break
        else:
            self.raw_error(
                'Error while listing languages in BOCA:\nNo language table found.'
            )
        table = typing.cast(Tag, table)
        rows = table.select('tr')
        languages: List[BocaLanguage] = []
        for row in rows[1:]:
            cells = row.select('td')
            index = int(cells[0].text.strip())
            name = cells[1].text.strip()
            extension = cells[2].text.strip()
            languages.append(BocaLanguage(index=index, name=name, extension=extension))
        return languages

    def list_runs(self, only_judged: bool = True) -> List[BocaRun]:
        _, html = self.open(
            f'{self.base_url}/admin/run.php'
            if not self.is_judge
            else f'{self.base_url}/judge/runchief.php',
            error_msg='Error while listing runs in BOCA',
        )

        soup = BeautifulSoup(html, 'html.parser')
        rows = soup.select('form[name="form1"] table tr')
        off = 1 if self.is_judge else 0

        runs: List[BocaRun] = []
        for row in rows[1:]:
            cells = row.select('td')

            run_number = cells[0].text.strip()
            site_number = cells[1].text.strip()
            shortname = cells[4 - off].text.strip()
            answer = cells[-1 - off].text.strip()
            time = int(cells[3 - off].text.strip())
            user = (
                cells[2].text.strip() if not self.is_judge else cells[-1].text.strip()
            )
            status = cells[-4 - off].text.strip().lower()

            if only_judged and status != 'judged':
                continue
            outcome = _parse_answer_as_outcome(answer)
            if outcome is None:
                if status == 'judged':
                    continue
            runs.append(
                BocaRun(
                    run_number=run_number,
                    site_number=site_number,
                    problem_shortname=shortname,
                    outcome=outcome,
                    time=time,
                    status=status,
                    user=user,
                )
            )

        return runs

    def wait_for_all_judged(self, step: int = 5):
        while True:
            runs = self.list_runs(only_judged=False)
            if all(run.outcome is not None for run in runs):
                return
            time.sleep(step)

    def retrieve_run(self, run: BocaRun) -> BocaDetailedRun:
        self.log(
            f'Retrieving run [item]{run.run_number}-{run.site_number}[/item] from BOCA...'
        )
        runedit_url = (
            f'{self.base_url}/admin/runedit.php'
            if not self.is_judge
            else f'{self.base_url}/judge/runeditchief.php'
        )
        url = (
            f'{runedit_url}?runnumber={run.run_number}&runsitenumber={run.site_number}'
        )
        _, html = self.open(
            url,
            error_msg=f'Error while downloading BOCA run [item]{run.run_number}-{run.site_number}[/item]',
        )

        soup = BeautifulSoup(html, 'html.parser')
        rows = soup.select('tr')

        href: Optional[str] = None
        filename: Optional[pathlib.Path] = None

        for row in rows:
            row_text = row.select('td')[0].text.strip().lower()
            if not row_text.endswith('code:'):
                continue
            link_col = row.select_one('td:nth-of-type(2) a:nth-of-type(1)')
            if link_col is None:
                continue
            href = str(link_col.attrs['href'])
            if filename is None:
                filename = pathlib.Path(link_col.text.strip())
            break

        if href is None or filename is None:
            self.raw_error(
                "Error while downloading run:\nNo link to team's code found."
            )

        link = self.br.find_link(url=href)
        tmp_file, _ = self.br.retrieve(link.absolute_url)
        if tmp_file is None:
            self.raw_error('Error while downloading run:\nDownloaded file is None.')

        return BocaDetailedRun(
            **run.model_dump(),
            language_repr='',
            code=pathlib.Path(tmp_file).read_text(),
            filename=filename,
            autojudge_answer='',
        )

    def retrieve_runs(self, only_judged: bool = True) -> List[BocaDetailedRun]:
        runs = self.list_runs(only_judged)
        return [self.retrieve_run(run) for run in runs]

    def download_run(
        self,
        run: BocaRun,
        into_dir: pathlib.Path,
        name: Optional[str] = None,
    ) -> pathlib.Path:
        detailed_run = self.retrieve_run(run)
        filename = detailed_run.filename.with_stem(
            name or f'{run.run_number}-{run.site_number}'
        )
        final_path = into_dir / filename
        final_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(detailed_run.code, final_path)
        return final_path

    def _set_starttime(
        self,
        form: mechanize.HTMLForm,
        start_time: datetime.datetime,
        tzinfo: datetime.tzinfo,
    ):
        start_time = start_time.astimezone(tzinfo)

        form['startdateh'] = f'{start_time.hour:02d}'
        form['startdatemin'] = f'{start_time.minute:02d}'
        form['startdated'] = f'{start_time.day:02d}'
        form['startdatem'] = f'{start_time.month:02d}'
        form['startdatey'] = f'{start_time.year}'

        self.log(
            f'Setting start time to {start_time.hour}:{start_time.minute} {start_time.day}/{start_time.month}/{start_time.year}'
        )

    def create_and_activate_contest(self):
        _, html = self.open(
            f'{self.base_url}/system/contest.php?new=1',
            error_msg='Error while creating contest in BOCA',
        )

        contest_page = self.get_redirect(html)
        self.log(f'Redirected to contest page: {contest_page}')
        if contest_page is None:
            self.pretty_print(html)
            self.raw_error(
                'Error while creating contest:\nNo redirect to contest page found.'
            )

        _, html = self.open(
            contest_page, error_msg='Error while opening the contest page'
        )

        try:
            self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Contest activation form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)
        form['confirmation'] = 'confirm'

        response = self.br.submit(name='Submit3', type='submit', nr=1)

        def alert_ok_fn(alert: str) -> bool:
            return 'You must log in the new contest' in alert

        self.log_response_alert(
            response,
            'Error while activating contest',
            alert_ok_fn=alert_ok_fn,
        )
        self.log('Contest activated successfully')

    def infer_timezone(self) -> datetime.tzinfo:
        _, html = self.open(
            f'{self.base_url}/admin/site.php',
            error_msg='Error while inferring timezone in BOCA',
        )

        match = START_DATE_REGEX.search(html)
        if match is None:
            self.raw_error(
                'Error while inferring timezone in BOCA:\nNo start date found.'
            )
        start_date = match.group(1)
        parsed_date = dateparser.parse(
            start_date, settings={'RETURN_AS_TIMEZONE_AWARE': True}
        )
        if parsed_date is None:
            self.raw_error(
                'Error while inferring timezone in BOCA:\nCould not parse start date.'
            )
        return parsed_date.tzinfo or datetime.timezone.utc

    def configure_contest(
        self,
        start_time: Optional[datetime.datetime] = None,
    ):
        tzinfo = self.infer_timezone()

        _, html = self.open(
            f'{self.base_url}/admin/contest.php',
            error_msg='Error while configuring contest in BOCA',
        )

        try:
            self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Contest activation form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)
        form['confirmation'] = 'confirm'

        if start_time is not None:
            self._set_starttime(form, start_time, tzinfo)

        req = self.br.click(name='Submit3', type='submit', nr=1)
        pretty_print_request_data(req)
        self.open(req)
        self.log('Contest configured successfully')

    def configure_main_site(
        self, autojudge: Optional[bool] = None, chief: Optional[str] = None
    ):
        _, html = self.open(
            f'{self.base_url}/admin/site.php',
            error_msg='Error while configuring main site in BOCA',
        )

        try:
            form = self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Main site configuration form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)
        form['confirmation'] = 'confirm'

        if autojudge is not None:
            self.br.find_control(name='autojudge', type='checkbox').items[
                0
            ].selected = autojudge

        if chief is not None:
            form['chiefname'] = chief

        req = self.br.click(name='Submit1', type='submit', nr=0)
        pretty_print_request_data(req)
        self.open(req)
        self.log('Main site configured successfully')

    def create_judge_account(self, password: str = 'boca'):
        _, html = self.open(
            f'{self.base_url}/admin/user.php',
            error_msg='Error while creating judge account in BOCA',
        )

        try:
            self.br.select_form(name='form3')
        except mechanize.FormNotFoundError:
            self.error(
                'Judge account creation form not found in BOCA website. This might happen when the login failed.'
            )

        salt = self.find_salt(html, 'form3.passwordo')
        console.console.print(f'Using salt [item]{salt}[/item]')
        admin_pwd_hash = self.hash(self.password, salt)

        self.br.form = self.br.global_form()
        form = typing.cast(mechanize.HTMLForm, self.br.form)

        form.method = 'POST'
        form.set_all_readonly(False)
        form['confirmation'] = 'confirm'
        form['usernumber'] = '42565759'
        form['username'] = 'judge'
        form['usertype'] = ['judge']
        form['usermultilogin'] = ['t']
        form['userfullname'] = 'Judge RBX'
        form['passwordn1'] = self.hash_two(password, self.password)
        form['passwordn2'] = self.hash_two(password, self.password)
        form['passwordo'] = admin_pwd_hash

        req = self.br.click(name='Submit', type='submit')
        pretty_print_request_data(req)
        self.open(req)
        self.log('Judge account created successfully')

    def wait_for_problem(self, problem_index_in_contest: int, timeout: int, step: int):
        _, html = self.open(
            f'{self.base_url}/judge/team.php',
            error_msg='Error while waiting for problem in BOCA',
        )

        self.log(
            f'Waiting for problem [item]{problem_index_in_contest}[/item] in BOCA (timeout: {timeout}s, step: {step}s)...'
        )
        soup = BeautifulSoup(html, 'html.parser')
        options = soup.select('select[name="problem"] option')
        available_indices = set(
            int(option.attrs['value'])
            for option in options
            if option.attrs.get('value') is not None
        )
        if problem_index_in_contest not in available_indices:
            if timeout <= 0:
                self.raw_error(
                    f'Problem index [item]{problem_index_in_contest}[/item] not found in BOCA.'
                )
            time.sleep(step)
            return self.wait_for_problem(problem_index_in_contest, timeout - step, step)
        return

    def submit_as_judge(
        self,
        problem_index_in_contest: int,
        language_index_in_contest: int,
        file: pathlib.Path,
        wait: int = 0,
    ):
        if wait > 0:
            self.wait_for_problem(problem_index_in_contest, wait, 5)
        _, html = self.open(
            f'{self.base_url}/judge/team.php',
            error_msg='Error while submitting problem to BOCA',
        )

        try:
            self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Judge submission form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)
        form['confirmation'] = 'confirm'

        form['problem'] = [str(problem_index_in_contest)]
        form['language'] = [str(language_index_in_contest)]

        with file.open('rb') as f:
            form.add_file(
                f,
                filename=file.name,
                name='sourcefile',
            )

            req = self.br.click(name='Submit', type='submit')
            pretty_print_request_data(req)
            self.open(req)
            self.log('Judge submission sent successfully')


@functools.cache
def get_boca_scraper(
    base_url: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> BocaScraper:
    return BocaScraper(base_url, username, password)


class ContestSnapshot:
    def __init__(
        self,
        problems: Optional[List[BocaProblem]] = None,
        languages: Optional[List[BocaLanguage]] = None,
        runs: Optional[List[BocaRun]] = None,
        detailed_runs: Optional[List[BocaDetailedRun]] = None,
    ):
        self.problems = problems or []
        self.languages = languages or []
        self.runs = runs or []
        self.detailed_runs = detailed_runs or []

    def __str__(self) -> str:
        return f'ContestSnapshot(problems={self.problems}, languages={self.languages})'

    def __repr__(self) -> str:
        return self.__str__()

    def get_problem_by_shortname(self, shortname: str) -> BocaProblem:
        for problem in self.problems:
            if problem.shortname == shortname:
                return problem
        raise ValueError(f'Problem with shortname {shortname} not found')

    def get_problem_by_basename(self, basename: str) -> BocaProblem:
        for problem in self.problems:
            if problem.basename == basename:
                return problem
        raise ValueError(f'Problem with basename {basename} not found')

    def get_problem_by_index(self, index: int) -> BocaProblem:
        for problem in self.problems:
            if problem.index == index:
                return problem
        raise ValueError(f'Problem with index {index} not found')

    def get_language_by_name(self, name: str) -> BocaLanguage:
        for language in self.languages:
            if language.name == name:
                return language
        raise ValueError(f'Language with name {name} not found')

    def get_language_by_extension(self, extension: str) -> BocaLanguage:
        for language in self.languages:
            if language.extension == extension:
                return language
        raise ValueError(f'Language with extension {extension} not found')

    def get_language_by_index(self, index: int) -> BocaLanguage:
        for language in self.languages:
            if language.index == index:
                return language
        raise ValueError(f'Language with index {index} not found')

    def get_run_by_number(self, number: int) -> BocaRun:
        for run in self.runs:
            if run.run_number == number:
                return run
        raise ValueError(f'Run with number {number} not found')

    def get_detailed_run_by_number(self, number: int) -> BocaDetailedRun:
        for run in self.detailed_runs:
            if run.run_number == number:
                return run
        raise ValueError(f'Run with number {number} not found')

    def get_detailed_run_by_path(self, path: pathlib.Path) -> BocaDetailedRun:
        for run in self.detailed_runs:
            if run.filename.name == path.name:
                return run
        raise ValueError(f'Run with path {path} not found')


def create_snapshot(
    scraper: BocaScraper, detailed_runs: bool = False
) -> ContestSnapshot:
    scraper.login()
    problems = scraper.list_problems()
    languages = scraper.list_languages()
    runs = scraper.list_runs()
    detailed = scraper.retrieve_runs() if detailed_runs else []
    return ContestSnapshot(problems, languages, runs, detailed)



import functools
import pathlib
import shutil
from typing import Dict, List, Optional, Set

import typer

from rbx import console
from rbx.box import checkers, package, testcase_utils, validators
from rbx.box.code import SanitizationLevel, compile_item, run_item
from rbx.box.schema import (
    CodeItem,
    GeneratorCall,
    TaskType,
    Testcase,
)
from rbx.box.tasks import run_solution_on_testcase
from rbx.box.testcase_extractors import (
    GenerationMetadata,
    GenerationTestcaseEntry,
    TestcaseGroupVisitor,
    extract_generation_testcases,
    run_testcase_visitor,
)
from rbx.box.testcase_utils import (
    TestcaseEntry,
    fill_output_for_defined_testcase,
    find_built_testcases,
)
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    Evaluation,
    Outcome,
)
from rbx.utils import StatusProgress


def _compile_generator(generator: CodeItem) -> str:
    return compile_item(generator, sanitized=SanitizationLevel.PREFER)


@functools.cache
def _warn_once_about_crlf():
    console.console.print(
        '[warning]It seems a few files have CRLF (\\r\\n) line endings.[/warning]'
    )
    console.console.print(
        '[warning]This usually happens when the file is created on Windows. Please convert the file to LF (\\n) line endings.[/warning]'
    )
    console.console.print(
        '[warning]If you are in VSCode, you can make sure LF (\\n) line endings are used by changing the [item]"files.eol"[/item] setting.[/warning]'
    )


@functools.cache
def _warn_about_crlf(path: pathlib.Path):
    _warn_once_about_crlf()
    console.console.print(
        f'[warning]Testcase file [item]{path}[/item] has CRLF (\\r\\n) line endings, converting to LF (\\n).[/warning]'
    )


def _check_crlf(path: pathlib.Path):
    should_fix = False
    with open(path, 'rb') as f:
        for line in f:
            if line.endswith(b'\r\n'):
                _warn_about_crlf(path)
                should_fix = True
                break

    if should_fix:
        path.write_text('\n'.join(path.read_text().splitlines()) + '\n')


def _copy_testcase_over(
    testcase: Testcase,
    dest: Testcase,
):
    testcase = fill_output_for_defined_testcase(testcase)
    dest.inputPath.parent.mkdir(parents=True, exist_ok=True)
    _check_crlf(testcase.inputPath)
    shutil.copy(
        str(testcase.inputPath),
        str(dest.inputPath),
    )
    if (
        testcase.outputPath is not None
        and testcase.outputPath.is_file()
        and dest.outputPath is not None
    ):
        _check_crlf(testcase.outputPath)
        dest.outputPath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(
            str(testcase.outputPath),
            str(dest.outputPath),
        )


def _copy_testcase_output_over(
    src_output_path: pathlib.Path,
    dest_output_path: pathlib.Path,
    suffix: str,
    dry_run: bool = False,
) -> bool:
    dest_output_path.parent.mkdir(parents=True, exist_ok=True)

    src_path = src_output_path.with_suffix(suffix)
    if not src_path.is_file():
        return False

    if dry_run:
        return True

    _check_crlf(src_path)
    shutil.copy(str(src_path), str(dest_output_path.with_suffix(suffix)))
    return True


def _copy_testcase_outputs_over(
    testcase: Testcase,
    dest: Testcase,
    pipes: bool = False,
    only_pipes: bool = False,
    dry_run: bool = False,
):
    if only_pipes:
        pipes = True
    assert dest.outputPath is not None
    if not dry_run:
        dest.outputPath.parent.mkdir(parents=True, exist_ok=True)

    has_copied = False

    if (
        not only_pipes
        and testcase.outputPath is not None
        and testcase.outputPath.is_file()
    ):
        if not dry_run:
            _check_crlf(testcase.outputPath)
            shutil.copy(str(testcase.outputPath), str(dest.outputPath))
        has_copied = True

    if not pipes:
        return has_copied

    reference_path = testcase.outputPath or testcase.inputPath
    if _copy_testcase_output_over(
        reference_path, dest.outputPath, '.pin', dry_run=dry_run
    ):
        has_copied = True

    if _copy_testcase_output_over(
        reference_path, dest.outputPath, '.pout', dry_run=dry_run
    ):
        has_copied = True

    if _copy_testcase_output_over(
        reference_path, dest.outputPath, '.pio', dry_run=dry_run
    ):
        has_copied = True

    return has_copied


def _needs_output(generation_entries: List[GenerationTestcaseEntry]) -> bool:
    for entry in generation_entries:
        tc = entry.metadata.copied_to
        if not tc.inputPath.is_file():
            continue
        if entry.metadata.copied_from is not None and _copy_testcase_outputs_over(
            entry.metadata.copied_from, tc, dry_run=True
        ):
            continue
        return True
    return False


def get_all_built_testcases() -> Dict[str, List[Testcase]]:
    pkg = package.find_problem_package_or_die()
    res = {group.name: find_built_testcases(group) for group in pkg.testcases}
    return res


def get_call_from_string(call_str: str) -> GeneratorCall:
    try:
        name, args = call_str.split(None, 1)
    except ValueError:
        return GeneratorCall(name=call_str, args='')
    return GeneratorCall(name=name, args=args)


async def _get_necessary_generators_for_groups(
    groups: Optional[Set[str]] = None,
) -> Set[str]:
    necessary_generators = set()

    class NecessaryGeneratorsVisitor(TestcaseGroupVisitor):
        async def visit(self, entry: GenerationTestcaseEntry):
            if entry.metadata.generator_call is not None:
                if (
                    package.get_generator_or_nil(entry.metadata.generator_call.name)
                    is None
                ):
                    console.console.print(
                        f'[error]Generator [item]{entry.metadata.generator_call.name}[/item] is not present in the package.[/error]'
                    )
                    if entry.metadata.generator_script is not None:
                        console.console.print(
                            f'[error]This generator is referenced from [item]{entry.metadata.generator_script}[/item].[/error]'
                        )
                    raise typer.Exit(1)
                necessary_generators.add(entry.metadata.generator_call.name)

    await run_testcase_visitor(NecessaryGeneratorsVisitor(groups))

    return necessary_generators


def compile_generators(
    tracked_generators: Set[str],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    def update_status(text: str):
        if progress is not None:
            progress.update(text)

    generator_to_compiled_digest = {}

    for generator_name in tracked_generators:
        generator = package.get_generator(generator_name)
        update_status(f'Compiling generator [item]{generator.name}[/item]')
        try:
            generator_to_compiled_digest[generator.name] = _compile_generator(generator)
        except:
            console.console.print(
                f'[error]Failed compiling generator [item]{generator.name}[/item].[/error]'
            )
            raise

    return generator_to_compiled_digest


def expand_generator_call(call: GeneratorCall) -> GeneratorCall:
    from rbx.box.stressing import generator_parser

    vars = package.find_problem_package_or_die().expanded_vars
    generator_for_args = generator_parser.Generator(vars)
    parsed_args = generator_parser.parse(call.args or '')
    return call.model_copy(update={'args': generator_for_args.generate(parsed_args)})


async def generate_standalone(
    spec: GenerationMetadata,
    validate: bool = True,
    group_entry: Optional[TestcaseEntry] = None,
    generator_digest: Optional[str] = None,
    validator_digest: Optional[str] = None,
    progress: Optional[StatusProgress] = None,
):
    def _print_error_header(text: Optional[str] = None):
        prefix = 'Failed generating test'
        if group_entry is not None:
            prefix += (
                f' [item]{group_entry.group}[/item]/[item]{group_entry.index}[/item]'
            )
        suffix = '.'
        if text:
            suffix = f': {text}'
        if spec.generator_call is not None:
            console.console.print(
                f'[error]{prefix} using generator call [info]{spec.generator_call.name} {spec.generator_call.args}[/info]{suffix}[/error]'
            )
        else:
            console.console.print(f'[error]{prefix}{suffix}[/error]')

    if spec.generator_call is not None:
        call = spec.generator_call

        generation_stderr = DigestHolder()

        # Get generator item
        generator = package.get_generator(call.name)
        if generator_digest is None:
            if progress:
                progress.update(f'Compiling generator {generator.name}...')
            generator_digest = _compile_generator(generator)

        if progress:
            progress.update(
                f'Generating testcase [status]{generator.name} {call.args}[/status]...'
            )
        generation_log = await run_item(
            generator,
            DigestOrSource.create(generator_digest),
            stdout=DigestOrDest.create(spec.copied_to.inputPath),
            stderr=DigestOrDest.create(generation_stderr),
            extra_args=call.args or None,
        )
        if not generation_log or generation_log.exitcode != 0:
            _print_error_header()
            if generation_log is not None:
                console.console.print(
                    f'[error]Summary:[/error] {generation_log.get_summary()}'
                )
            if generation_stderr.value is not None:
                console.console.print('[error]Stderr:[/error]')
                console.console.print(
                    package.get_digest_as_string(generation_stderr.value) or ''
                )

            raise typer.Exit(1)
    elif spec.copied_from is not None:
        _copy_testcase_over(spec.copied_from, spec.copied_to)

    validator = package.get_validator_or_nil()
    # Run validator, if it is available.
    if validator is not None and validate:
        if validator_digest is None:
            if progress:
                progress.update('Compiling validator...')
            validator_tp = validators.compile_main_validator()
            assert validator_tp is not None
            _, validator_digest = validator_tp
        if progress:
            progress.update('Validating test...')
        validation_info = await validators.validate_one_off(
            spec.copied_to.inputPath,
            validator,
            validator_digest,
        )
        if not validation_info.ok:
            _print_error_header('failed validating testcase.')
            console.console.print(f'[error]Message:[/error] {validation_info.message}')
            console.console.print(
                f'Testcase written at [item]{spec.copied_to.inputPath}[/item]'
            )
            raise typer.Exit(1)


async def generate_testcases(
    progress: Optional[StatusProgress] = None, groups: Optional[Set[str]] = None
):
    def step():
        if progress is not None:
            progress.step()

    compiled_generators = compile_generators(
        progress=progress,
        tracked_generators=await _get_necessary_generators_for_groups(groups),
    )

    testcase_utils.clear_built_testcases()

    class BuildTestcaseVisitor(TestcaseGroupVisitor):
        async def visit(self, entry: GenerationTestcaseEntry):
            if entry.metadata.copied_from is not None:
                _copy_testcase_over(
                    entry.metadata.copied_from,
                    entry.metadata.copied_to,
                )

            if entry.metadata.generator_call is not None:
                await generate_standalone(
                    entry.metadata,
                    group_entry=entry.group_entry,
                    validate=False,
                    generator_digest=compiled_generators[
                        entry.metadata.generator_call.name
                    ],
                )
            step()

    await run_testcase_visitor(BuildTestcaseVisitor(groups))


async def generate_output_for_testcase(
    model_solution: CodeItem,
    model_solution_digest: str,
    testcase: Testcase,
    interactor_digest: Optional[str] = None,
    capture_pipes: Optional[bool] = None,
):
    assert testcase.outputPath is not None
    testcase.inputPath.parent.mkdir(parents=True, exist_ok=True)
    testcase.outputPath.parent.mkdir(parents=True, exist_ok=True)

    eval: Evaluation = await run_solution_on_testcase(
        model_solution,
        model_solution_digest,
        None,
        testcase,
        interactor_digest=interactor_digest,
        use_retries=False,
        use_timelimit=False,
        capture_pipes=capture_pipes,
    )

    if eval.result.outcome.is_slow() and eval.result.no_tle_outcome == Outcome.ACCEPTED:
        console.console.print(
            f'[warning]Testcase [item]{testcase.inputPath}[/item] finished in TLE, but test was generated successfully.[/warning]'
        )
    elif eval.result.outcome != Outcome.ACCEPTED:
        console.console.print(
            f'[error]Failed generating output for [item]{testcase.inputPath}[/item][/error]',
        )
        console.console.print(f'[error]Summary:[/error] {eval.log.get_summary()}')
        console.console.print(
            f'[warning]Verdict: [item]{eval.result.outcome.value}[/item][/warning]',
        )
        console.console.print(
            f'[warning]Message: [info]{eval.result.message}[/info][/warning]',
        )
        console.console.print(f'Input written at [item]{testcase.inputPath}[/item]')
        console.console.print(f'Output written at [item]{testcase.outputPath}[/item]')
        if eval.log.stderr_absolute_path is not None:
            console.console.print(
                f'Stderr written at [item]{eval.log.stderr_absolute_path}[/item]'
            )

        raise typer.Exit(1)


async def generate_outputs_for_testcases(
    entries: List[TestcaseEntry],
    progress: Optional[StatusProgress] = None,
):
    def step():
        if progress is not None:
            progress.step()

    generation_entries = await extract_generation_testcases(entries)
    needs_output = _needs_output(generation_entries)

    main_solution = package.get_main_solution()
    solution_digest_map = {}

    pkg = package.find_problem_package_or_die()

    if pkg.type == TaskType.COMMUNICATION and needs_output:
        interactor_digest = checkers.compile_interactor(progress)
    else:
        interactor_digest = None

    if main_solution is not None and needs_output:
        if progress:
            progress.update('Compiling main solution...')
        try:
            solution_digest_map[main_solution.path] = compile_item(main_solution)
        except:
            console.console.print('[error]Failed compiling main solution.[/error]')
            raise

    for entry in generation_entries:
        if (
            entry.model_solution is not None
            and entry.model_solution.path not in solution_digest_map
        ):
            if progress:
                progress.update(
                    f'Compiling model solution [item]{entry.model_solution.path}[/item]...'
                )
            try:
                solution_digest_map[entry.model_solution.path] = compile_item(
                    entry.model_solution
                )
            except:
                console.console.print(
                    f'[error]Failed compiling model solution [item]{entry.model_solution.path}[/item].[/error]'
                )
                raise

    gen_runs_dir = package.get_problem_runs_dir() / '.gen'
    shutil.rmtree(str(gen_runs_dir), ignore_errors=True)
    gen_runs_dir.mkdir(parents=True, exist_ok=True)

    for entry in generation_entries:
        tc = entry.metadata.copied_to
        if not tc.inputPath.is_file():
            return
        assert tc.outputPath is not None

        if entry.metadata.copied_from is not None and _copy_testcase_outputs_over(
            entry.metadata.copied_from, tc
        ):
            # Copy remaining pipe files.
            _copy_testcase_outputs_over(entry.metadata.copied_from, tc, pipes=True)
            step()
            continue

        assert needs_output
        model_solution = entry.model_solution or main_solution
        if (
            model_solution is None or model_solution.path not in solution_digest_map
        ) and not tc.outputPath.is_file():
            console.console.print(
                '[error]No main/model solution found to generate outputs for testcases.[/error]',
            )
            raise typer.Exit(1)

        assert model_solution is not None
        model_solution_digest = solution_digest_map[model_solution.path]
        capture_pipes = None
        if (
            pkg.type == TaskType.COMMUNICATION
            and entry.metadata.copied_from is not None
        ):
            # If some pipe file is already specified, we don't need to capture the pipes
            # when running the program.
            capture_pipes = not _copy_testcase_outputs_over(
                entry.metadata.copied_from, tc, only_pipes=True, dry_run=True
            )

        await generate_output_for_testcase(
            model_solution,
            model_solution_digest,
            tc,
            interactor_digest=interactor_digest,
            capture_pipes=capture_pipes,
        )
        if entry.metadata.copied_from is not None:
            # Copy remaining pipe files.
            _copy_testcase_outputs_over(entry.metadata.copied_from, tc, pipes=True)
        step()



import pathlib

from rbx import utils
from rbx.box.presets.schema import Preset, TrackedAsset, Tracking
from rbx.box.testing.testing_shared import PathOrStr, TestingShared


class TestingPreset(TestingShared):
    def __init__(self, root: PathOrStr):
        super().__init__(root)
        self._yml = None

    def initialize(self):
        if not self.yml_path.exists():
            self.yml_path.parent.mkdir(parents=True, exist_ok=True)
            self.yml_path.touch()
            self.yml_path.write_text(
                utils.model_to_yaml(
                    Preset(
                        name='test-preset',
                        uri='rsalesc/test-preset',
                        env=pathlib.Path('env.rbx.yml'),
                        tracking=Tracking(),  # Explicitly include tracking
                    )
                )
            )
            self.add_from_resources(
                pathlib.Path('env.rbx.yml'), pathlib.Path('presets/default/env.rbx.yml')
            )

    @property
    def yml_path(self) -> pathlib.Path:
        return self.root / 'preset.rbx.yml'

    @property
    def yml(self) -> Preset:
        if self._yml is None:
            self._yml = utils.model_from_yaml(Preset, self.yml_path.read_text())
        return self._yml

    def save(self):
        self.yml_path.write_text(utils.model_to_yaml(self.yml))
        self._yml = None

    def set_name(self, name: str):
        """Set the preset name."""
        self.yml.name = name
        self.save()

    def set_uri(self, uri: str):
        """Set the preset URI."""
        self.yml.uri = uri
        self.save()

    def set_env(self, env_path: PathOrStr):
        """Set the environment file path."""
        self.yml.env = pathlib.Path(env_path)
        self.save()

    def set_problem_path(self, path: PathOrStr):
        """Set the problem package path."""
        self.yml.problem = pathlib.Path(path)
        self.save()

    def set_contest_path(self, path: PathOrStr):
        """Set the contest package path."""
        self.yml.contest = pathlib.Path(path)
        self.save()

    def add_problem_tracked_asset(self, path: PathOrStr, symlink: bool = False):
        """Add a tracked asset to the problem tracking list."""
        # Create a new tracking object with the updated problem list
        current_tracking = self.yml.tracking
        new_problem_list = current_tracking.problem + [
            TrackedAsset(path=pathlib.Path(path), symlink=symlink)
        ]
        self.yml.tracking = Tracking(
            problem=new_problem_list, contest=current_tracking.contest
        )
        self.save()

    def add_contest_tracked_asset(self, path: PathOrStr, symlink: bool = False):
        """Add a tracked asset to the contest tracking list."""
        # Create a new tracking object with the updated contest list
        current_tracking = self.yml.tracking
        new_contest_list = current_tracking.contest + [
            TrackedAsset(path=pathlib.Path(path), symlink=symlink)
        ]
        self.yml.tracking = Tracking(
            problem=current_tracking.problem, contest=new_contest_list
        )
        self.save()

    def create_problem_package(self):
        """Create a basic problem package structure."""
        if self.yml.problem:
            problem_dir = self.root / self.yml.problem
            problem_dir.mkdir(parents=True, exist_ok=True)

            # Create a basic problem.rbx.yml
            problem_yml = problem_dir / 'problem.rbx.yml'
            if not problem_yml.exists():
                problem_yml.write_text("""---
name: "test-problem"
timeLimit: 1000
memoryLimit: 256
""")

    def create_contest_package(self):
        """Create a basic contest package structure."""
        if self.yml.contest:
            contest_dir = self.root / self.yml.contest
            contest_dir.mkdir(parents=True, exist_ok=True)

            # Create a basic contest.rbx.yml
            contest_yml = contest_dir / 'contest.rbx.yml'
            if not contest_yml.exists():
                contest_yml.write_text("""---
name: "Test Contest"
duration: 180
""")

    def create_symlink(self, link_path: PathOrStr, target_path: PathOrStr):
        """Create a symlink from link_path to target_path relative to the preset root."""
        link = self.root / link_path
        target = pathlib.Path(target_path)

        # Ensure parent directory exists
        link.parent.mkdir(parents=True, exist_ok=True)

        # Remove existing file/symlink if present
        if link.exists() or link.is_symlink():
            link.unlink()

        # Create symlink
        link.symlink_to(target)

    def verify_file_exists(self, path: PathOrStr) -> bool:
        """Verify that a file exists in the preset."""
        return (self.root / path).exists()

    def verify_symlink(self, path: PathOrStr, expected_target: PathOrStr) -> bool:
        """Verify that a symlink exists and points to the expected target."""
        link = self.root / path
        if not link.is_symlink():
            return False
        return link.readlink() == pathlib.Path(expected_target)

    def get_file_content(self, path: PathOrStr) -> str:
        """Get the content of a file in the preset."""
        return (self.root / path).read_text()

    def get_problem_dir(self) -> pathlib.Path:
        """Get the problem package directory."""
        if not self.yml.problem:
            raise ValueError('No problem package defined in preset')
        return self.root / self.yml.problem

    def get_contest_dir(self) -> pathlib.Path:
        """Get the contest package directory."""
        if not self.yml.contest:
            raise ValueError('No contest package defined in preset')
        return self.root / self.yml.contest



import pathlib
import tempfile
from typing import Optional

import typer

from rbx import console
from rbx.box import builder, cd, package
from rbx.box.environment import VerificationLevel
from rbx.box.packaging.boca.packager import BocaPackager
from rbx.box.packaging.importer import BaseImporter
from rbx.box.packaging.moj.packager import MojPackager
from rbx.box.packaging.packager import BasePackager, BuiltStatement
from rbx.box.packaging.polygon.importer import PolygonImporter
from rbx.box.packaging.polygon.packager import PolygonPackager
from rbx.box.statements.build_statements import build_statement

PACKAGER_REGISTRY = {
    'polygon': PolygonPackager,
    'boca': BocaPackager,
    'moj': MojPackager,
}

IMPORTER_REGISTRY = {
    'polygon': PolygonImporter,
}


def get_packager(source: str, **kwargs) -> BasePackager:
    if source not in PACKAGER_REGISTRY:
        console.console.print(f'Unknown packager: {source}')
        raise typer.Exit(1)
    return PACKAGER_REGISTRY[source](**kwargs)


def get_importer(source: str, **kwargs) -> BaseImporter:
    if source not in IMPORTER_REGISTRY:
        console.console.print(f'Unknown importer: {source}')
        raise typer.Exit(1)
    return IMPORTER_REGISTRY[source](**kwargs)


async def convert(
    pkg_dir: pathlib.Path,
    into_dir: pathlib.Path,
    source: str,
    destination: str,
    main_language: Optional[str] = None,
) -> pathlib.Path:
    importer = get_importer(source, main_language=main_language)
    packager = get_packager(destination)
    await importer.import_package(pkg_dir, into_dir)

    with cd.new_package_cd(into_dir):
        package.clear_package_cache()

        pkg = package.find_problem_package_or_die()

        if not await builder.build(VerificationLevel.NONE.value):
            console.console.print('[error]Failed to build the problem.[/error]')
            raise typer.Exit(1)

        built_statements = []
        for statement_type in packager.statement_types():
            for language in packager.languages():
                statement = packager.get_statement_for_language(language)
                statement_path = build_statement(statement, pkg, statement_type)
                built_statements.append(
                    BuiltStatement(statement, statement_path, statement_type)
                )

        with tempfile.TemporaryDirectory() as td:
            result_path = packager.package(
                package.get_build_path(), pathlib.Path(td), built_statements
            )
            return result_path



import abc
import asyncio
import dataclasses
import io
import json
import logging
import os
import pathlib
import select
import stat
import subprocess
import sys
import typing
from typing import IO, Any, Dict, List, Optional, Tuple

import pydantic

from rbx import utils
from rbx.grading.judge import cacher, storage

logger = logging.getLogger(__name__)

MERGE_STDERR = pathlib.PosixPath('/dev/stdout')


# Thread-safe version of asyncio.Event.
class Event_ts(asyncio.Event):
    def __init__(self):
        super().__init__()
        self._inherited_loop = asyncio.get_event_loop()

    def get_loop(self):
        if self._inherited_loop is None:
            return None
        if self._inherited_loop.is_closed():
            return None
        return self._inherited_loop

    def set_loop(self, loop):
        self._inherited_loop = loop

    def set(self):
        loop = self.get_loop()
        if loop is not None:
            loop.call_soon_threadsafe(super().set)

    def clear(self):
        loop = self.get_loop()
        if loop is not None:
            loop.call_soon_threadsafe(super().clear)


def wait_without_std(
    procs: List[subprocess.Popen], actually_pipe_to_stdout: bool = False
) -> List[int]:
    """Wait for the conclusion of the processes in the list, avoiding
    starving for input and output.

    procs (list): a list of processes as returned by Popen.

    return (list): a list of return codes.

    """

    def get_to_consume():
        """Amongst stdout and stderr of list of processes, find the
        ones that are alive and not closed (i.e., that may still want
        to write to).

        return (list): a list of open streams.

        """
        to_consume = []
        for process in procs:
            if process.poll() is None:  # If the process is alive.
                if process.stdout and not process.stdout.closed:
                    to_consume.append(process.stdout)
                if process.stderr and not process.stderr.closed:
                    to_consume.append(process.stderr)
        return to_consume

    # Close stdin; just saying stdin=None isn't ok, because the
    # standard input would be obtained from the application stdin,
    # that could interfere with the child process behaviour
    for process in procs:
        if process.stdin:
            process.stdin.close()

    # Read stdout and stderr to the end without having to block
    # because of insufficient buffering (and without allocating too
    # much memory). Unix specific.
    to_consume = get_to_consume()

    while len(to_consume) > 0:
        to_read = select.select(to_consume, [], [], 1.0)[0]
        for file_ in to_read:
            consumed = file_.read(8 * 1024)
            if actually_pipe_to_stdout:
                sys.stdout.buffer.write(consumed)
                sys.stdout.buffer.flush()
        to_consume = get_to_consume()

    return [process.wait() for process in procs]


@dataclasses.dataclass
class DirectoryMount:
    src: pathlib.Path
    dst: pathlib.Path
    options: Optional[str] = None


class SandboxParams(pydantic.BaseModel):
    """Parameters for the sandbox.

    box_id (int): the id of the sandbox.
    fsize (int|None): maximum file size.
    cgroup (bool): whether to use cgroups.
    dirs ([string]): directories to mount.
    preserve_env (bool): whether to preserve the environment.
    inherit_env ([string]): environment variables to inherit.
    set_env (Dict[string, string]): environment variables to set.
    verbosity (int): verbosity level.
    max_processes (int): maximum number of processes.

    """

    fsize: Optional[int] = None  # KiB
    cgroup: bool = False
    dirs: List[DirectoryMount] = []
    preserve_env: bool = False
    inherit_env: List[str] = []
    set_env: Dict[str, str] = {}
    verbosity: int = 0
    max_processes: Optional[int] = 1

    stdin_file: Optional[pathlib.Path] = None
    stdout_file: Optional[pathlib.Path] = None
    stderr_file: Optional[pathlib.Path] = None
    stack_space: Optional[int] = None  # MiB
    address_space: Optional[int] = None  # MiB
    timeout: Optional[int] = None  # ms
    wallclock_timeout: Optional[int] = None  # ms
    extra_timeout: Optional[int] = None  # ms
    reverse_io: bool = False
    pgid: Optional[int] = None

    def get_cacheable_params(self) -> Dict[str, Any]:
        return self.model_dump(mode='json', exclude_unset=True, exclude_none=True)

    def set_stdio(
        self,
        stdin: Optional[pathlib.Path] = None,
        stdout: Optional[pathlib.Path] = None,
    ):
        """Set the standard input/output files.

        stdin (Path): standard input file.
        stdout (Path): standard output file.

        """
        self.stdin_file = stdin
        self.stdout_file = stdout

    def set_stdall(
        self,
        stdin: Optional[pathlib.Path] = None,
        stdout: Optional[pathlib.Path] = None,
        stderr: Optional[pathlib.Path] = None,
    ):
        """Set the standard input/output/error files.

        stdin (Path): standard input file.
        stdout (Path): standard output file.
        stderr (Path): standard error file.

        """
        self.stdin_file = stdin
        self.stdout_file = stdout
        self.stderr_file = stderr

    def add_mapped_directory(
        self,
        src: pathlib.Path,
        dest: Optional[pathlib.Path] = None,
        options: Optional[str] = None,
        ignore_if_not_existing: bool = False,
    ):
        """Add src to the directory to be mapped inside the sandbox.

        src (Path): directory to make visible.
        dest (Path|None): if not None, the path where to bind src.
        options (str|None): if not None, isolate's directory rule options.
        ignore_if_not_existing (bool): if True, ignore the mapping when src
            does not exist (instead of having isolate terminate with an
            error).

        """
        if dest is None:
            dest = src
        if ignore_if_not_existing and not src.exists():
            return
        self.dirs.append(DirectoryMount(src, dest, options))


class SandboxLog(pydantic.BaseModel):
    """A log of the sandbox."""

    params: SandboxParams
    execution_time: float  # seconds
    memory_used: int  # bytes
    exitcode: int
    exitstatus: str
    killing_signal: Optional[int] = None
    exit_index: int = 0
    other_logs: Dict[str, Any] = pydantic.Field(default_factory=dict)

    def dump_other_logs(self) -> str:
        return json.dumps(self.other_logs)


class SandboxBase(abc.ABC):
    """A base class for all sandboxes, meant to contain common
    resources.

    """

    EXIT_SANDBOX_ERROR = 'sandbox error'
    EXIT_OK = 'ok'
    EXIT_TERMINATED = 'terminated'
    EXIT_SIGNAL = 'signal'
    EXIT_TIMEOUT = 'timeout'
    EXIT_TIMEOUT_WALL = 'wall timeout'
    EXIT_NONZERO_RETURN = 'nonzero return'
    EXIT_MEMORY_LIMIT_EXCEEDED = 'memory limit exceeded'
    EXIT_OUTPUT_LIMIT_EXCEEDED = 'output limit exceeded'

    file_cacher: cacher.FileCacher
    name: str
    temp_dir: Optional[pathlib.Path]

    def __init__(
        self,
        file_cacher: Optional[cacher.FileCacher] = None,
        name: Optional[str] = None,
        temp_dir: Optional[pathlib.Path] = None,
    ):
        """Initialization.

        file_cacher (FileCacher): an instance of the FileCacher class
            (to interact with FS), if the sandbox needs it.
        name (string|None): name of the sandbox, which might appear in the
            path and in system logs.
        temp_dir (Path|None): temporary directory to use; if None, use the
            default temporary directory.

        """
        self.file_cacher = file_cacher or cacher.FileCacher(storage.NullStorage())
        self.name = name if name is not None else 'unnamed'
        self.temp_dir = temp_dir
        self.cmd_file = pathlib.PosixPath('__commands__.log')

    @abc.abstractmethod
    def get_root_path(self) -> pathlib.Path:
        """Return the toplevel path of the sandbox.

        return (Path): the root path.

        """
        pass

    def use_soft_timeout(self) -> bool:
        return False

    def relative_path(self, path: pathlib.Path) -> pathlib.Path:
        """Translate from a relative path inside the sandbox to a
        system path.

        path (Path): relative path of the file inside the sandbox.

        return (string): the absolute path.

        """
        return self.get_root_path() / path

    def create_file(
        self, path: pathlib.Path, executable: bool = False, override: bool = False
    ) -> IO[bytes]:
        """Create an empty file in the sandbox and open it in write
        binary mode.

        path (Path): relative path of the file inside the sandbox.
        executable (bool): to set permissions.

        return (file): the file opened in write binary mode.

        """
        if executable:
            logger.debug('Creating executable file %s in sandbox.', path)
        else:
            logger.debug('Creating plain file %s in sandbox.', path)
        real_path = self.relative_path(path)
        if override:
            real_path.unlink(missing_ok=True)
        # Ensure directory exists.
        real_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            file_fd = os.open(str(real_path), os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            file_ = open(file_fd, 'wb')
        except OSError as e:
            logger.error(
                'Failed create file %s in sandbox. Unable to '
                'evalulate this submission. This may be due to '
                'cheating. %s',
                real_path,
                e,
                exc_info=True,
            )
            raise
        mod = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
        if executable:
            mod |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
        os.chmod(str(real_path), mod)
        return file_

    def create_symlink(
        self, path: pathlib.Path, from_path: pathlib.Path, override: bool = False
    ) -> Optional[pathlib.Path]:
        real_path = self.relative_path(path)
        if override:
            real_path.unlink(missing_ok=True)
        try:
            real_path.parent.mkdir(parents=True, exist_ok=True)
            real_path.symlink_to(utils.abspath(from_path))
        except NotImplementedError:
            return None
        return real_path

    def create_fifo(self, path: pathlib.Path, override: bool = False):
        real_path = self.relative_path(path)
        if override:
            real_path.unlink(missing_ok=True)
        os.mkfifo(str(real_path))
        return real_path

    def create_file_from_storage(
        self,
        path: pathlib.Path,
        digest: str,
        executable: bool = False,
        override: bool = False,
        try_symlink: bool = False,
    ):
        """Write a file taken from FS in the sandbox.

        path (Path): relative path of the file inside the sandbox.
        digest (string): digest of the file in FS.
        executable (bool): to set permissions.

        """
        if try_symlink and executable:
            symlink_path = self.file_cacher.transient_path_for_symlink(digest)
            if symlink_path is not None:
                created = self.create_symlink(
                    path,
                    from_path=symlink_path,
                    override=override,
                )
                if created is not None:
                    created.chmod(0o755)
                    return
        with self.create_file(path, executable, override=override) as dest_fobj:
            self.file_cacher.get_file_to_fobj(digest, dest_fobj)

    def create_file_from_bytes(
        self,
        path: pathlib.Path,
        content: bytes,
        executable: bool = False,
        override: bool = False,
    ):
        """Write some data to a file in the sandbox.

        path (Path): relative path of the file inside the sandbox.
        content (bytes): what to write in the file.
        executable (bool): to set permissions.

        """
        with self.create_file(path, executable, override=override) as dest_fobj:
            dest_fobj.write(content)

    def create_file_from_other_file(
        self,
        path: pathlib.Path,
        from_path: pathlib.Path,
        executable: bool = False,
        override: bool = False,
        try_symlink: bool = False,
    ):
        """Write a file taken from FS in the sandbox.

        path (Path): relative path of the file inside the sandbox.
        digest (string): digest of the file in FS.
        executable (bool): to set permissions.

        """
        if try_symlink and executable:
            created = self.create_symlink(
                path,
                from_path,
                override=override,
            )
            if created is not None:
                created.chmod(0o755)
                return
        with self.create_file(path, executable, override=override) as dest_fobj:
            with from_path.open('rb') as src_fobj:
                storage.copyfileobj(src_fobj, dest_fobj)

    def create_file_from_string(
        self,
        path: pathlib.Path,
        content: str,
        executable: bool = False,
        override: bool = False,
    ):
        """Write some data to a file in the sandbox.

        path (Path): relative path of the file inside the sandbox.
        content (str): what to write in the file.
        executable (bool): to set permissions.

        """
        return self.create_file_from_bytes(
            path, content.encode('utf-8'), executable, override=override
        )

    def get_file(
        self, path: pathlib.Path, trunc_len: Optional[int] = None
    ) -> IO[bytes]:
        """Open a file in the sandbox given its relative path.

        path (Path): relative path of the file inside the sandbox.
        trunc_len (int|None): if None, does nothing; otherwise, before
            returning truncate it at the specified length.

        return (file): the file opened in read binary mode.

        """
        logger.debug(f'Retrieving file {path} from sandbox.')
        real_path = self.relative_path(path)
        file_ = real_path.open('rb')
        if trunc_len is not None:
            file_ = Truncator(file_, trunc_len)
        return typing.cast(IO[bytes], file_)

    def get_file_text(
        self, path: pathlib.Path, trunc_len: Optional[int] = None
    ) -> IO[str]:
        """Open a file in the sandbox given its relative path, in text mode.

        Assumes encoding is UTF-8. The caller must handle decoding errors.

        path (Path): relative path of the file inside the sandbox.
        trunc_len (int|None): if None, does nothing; otherwise, before
            returning truncate it at the specified length.

        return (file): the file opened in read binary mode.

        """
        logger.debug('Retrieving text file %s from sandbox.', path)
        real_path = self.relative_path(path)
        file_ = real_path.open('rt', encoding='utf-8')
        if trunc_len is not None:
            file_ = Truncator(file_, trunc_len)
        return typing.cast(IO[str], file_)

    def get_file_to_bytes(
        self, path: pathlib.Path, maxlen: Optional[int] = 1024
    ) -> bytes:
        """Return the content of a file in the sandbox given its
        relative path.

        path (Path): relative path of the file inside the sandbox.
        maxlen (int): maximum number of bytes to read, or None if no
            limit.

        return (bytes): the content of the file up to maxlen bytes.

        """
        with self.get_file(path) as file_:
            if maxlen is None:
                return file_.read()
            else:
                return file_.read(maxlen)

    def get_file_to_string(
        self, path: pathlib.Path, maxlen: Optional[int] = 1024
    ) -> str:
        """Return the content of a file in the sandbox given its
        relative path.

        path (Path): relative path of the file inside the sandbox.
        maxlen (int): maximum number of bytes to read, or None if no
            limit.

        return (string): the content of the file up to maxlen bytes.

        """
        return self.get_file_to_bytes(path, maxlen).decode('utf-8')

    def get_file_to_storage(
        self,
        path: pathlib.Path,
        metadata: Optional[Dict[str, pydantic.BaseModel]] = None,
        trunc_len: Optional[int] = None,
    ) -> str:
        """Put a sandbox file in FS and return its digest.

        path (Path): relative path of the file inside the sandbox.
        metadata (Dict[str, pydantic.BaseModel]): the metadata for FS.
        trunc_len (int|None): if None, does nothing; otherwise, before
            returning truncate it at the specified length.

        return (str): the digest of the file.

        """
        with self.get_file(path, trunc_len=trunc_len) as file_:
            return self.file_cacher.put_file_from_fobj(file_, metadata)

    def stat_file(self, path: pathlib.Path) -> os.stat_result:
        """Return the stats of a file in the sandbox.

        path (Path): relative path of the file inside the sandbox.

        return (stat_result): the stat results.

        """
        return self.relative_path(path).stat()

    def file_exists(self, path: pathlib.Path) -> bool:
        """Return if a file exists in the sandbox.

        path (Path): relative path of the file inside the sandbox.

        return (bool): if the file exists.

        """
        return self.relative_path(path).exists()

    def remove_file(self, path: pathlib.Path):
        """Delete a file in the sandbox.

        path (Path): relative path of the file inside the sandbox.

        """
        self.relative_path(path).unlink(missing_ok=True)

    def glob(self, glob_expr: str) -> List[pathlib.Path]:
        return [
            path.relative_to(self.get_root_path())
            for path in self.get_root_path().glob(glob_expr)
        ]

    @abc.abstractmethod
    def run(
        self,
        command: List[str],
        params: SandboxParams,
    ) -> SandboxLog:
        pass

    @abc.abstractmethod
    def run_communication(
        self,
        command: List[str],
        params: SandboxParams,
        interactor_command: List[str],
        interactor_params: SandboxParams,
        merged_capture: Optional[pathlib.Path] = None,
    ) -> Tuple[SandboxLog, SandboxLog]:
        pass

    @abc.abstractmethod
    def initialize(self):
        """Initialize the sandbox.

        To be called at the beginning of the execution.

        """
        pass

    def reset(self):
        """Reset the sandbox.

        To be called at the beginning of the execution.

        """
        self.cleanup(delete=True)
        self.initialize()

    @abc.abstractmethod
    def cleanup(self, delete: bool = False):
        """Cleanup the sandbox.

        To be called at the end of the execution, regardless of
        whether the sandbox should be deleted or not.

        delete (bool): if True, also delete get_root_path() and everything it
            contains.

        """
        pass


class Truncator(io.RawIOBase):
    """Wrap a file-like object to simulate truncation.

    This file-like object provides read-only access to a limited prefix
    of a wrapped file-like object. It provides a truncated version of
    the file without ever touching the object on the filesystem.

    This class is only able to wrap binary streams as it relies on the
    readinto method which isn't provided by text (unicode) streams.

    """

    def __init__(self, fobj, size):
        """Wrap fobj and give access to its first size bytes.

        fobj (fileobj): a file-like object to wrap.
        size (int): the number of bytes that will be accessible.

        """
        self.fobj = fobj
        self.size = size

    def close(self):
        """See io.IOBase.close."""
        self.fobj.close()

    @property
    def closed(self):
        """See io.IOBase.closed."""
        return self.fobj.closed

    def readable(self):
        """See io.IOBase.readable."""
        return True

    def seekable(self):
        """See io.IOBase.seekable."""
        return True

    def readinto(self, b):
        """See io.RawIOBase.readinto."""
        # This is the main "trick": we clip (i.e. mask, reduce, slice)
        # the given buffer so that it doesn't overflow into the area we
        # want to hide (that is, out of the prefix) and then we forward
        # it to the wrapped file-like object.
        b = memoryview(b)[: max(0, self.size - self.fobj.tell())]
        return self.fobj.readinto(b)

    def seek(self, offset, whence=io.SEEK_SET):
        """See io.IOBase.seek."""
        # We have to catch seeks relative to the end of the file and
        # adjust them to the new "imposed" size.
        if whence == io.SEEK_END:
            if self.fobj.seek(0, io.SEEK_END) > self.size:
                self.fobj.seek(self.size, io.SEEK_SET)
            return self.fobj.seek(offset, io.SEEK_CUR)
        else:
            return self.fobj.seek(offset, whence)

    def tell(self):
        """See io.IOBase.tell."""
        return self.fobj.tell()

    def write(self, _):
        """See io.RawIOBase.write."""
        raise io.UnsupportedOperation('write')



import dataclasses
import pathlib
import shutil
import tempfile
from abc import ABC, abstractmethod
from typing import List, Tuple, Type

import typer

from rbx import console
from rbx.box import environment, header, limits_info, naming, package
from rbx.box.contest import contest_package
from rbx.box.contest.schema import ContestProblem, ContestStatement
from rbx.box.formatting import href
from rbx.box.generators import get_all_built_testcases
from rbx.box.schema import Package, TaskType, Testcase, TestcaseGroup
from rbx.box.statements.build_statements import build_statement
from rbx.box.statements.schema import Statement, StatementType


@dataclasses.dataclass
class BuiltStatement:
    statement: Statement
    path: pathlib.Path
    output_type: StatementType


@dataclasses.dataclass
class BuiltContestStatement:
    statement: ContestStatement
    path: pathlib.Path
    output_type: StatementType


@dataclasses.dataclass
class BuiltProblemPackage:
    path: pathlib.Path
    package: Package
    problem: ContestProblem


class BasePackager(ABC):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass

    @classmethod
    def task_types(cls) -> List[TaskType]:
        return [TaskType.BATCH]

    def languages(self):
        pkg = package.find_problem_package_or_die()

        res = set()
        for statement in pkg.expanded_statements:
            res.add(statement.language)
        return list(res)

    def package_basename(self):
        pkg = package.find_problem_package_or_die()
        shortname = naming.get_problem_shortname()
        if shortname is not None:
            return f'{shortname}-{pkg.name}'
        return pkg.name

    def statement_types(self) -> List[StatementType]:
        return [StatementType.PDF]

    @abstractmethod
    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ) -> pathlib.Path:
        pass

    # Helper methods.
    def get_built_testcases_per_group(self):
        return get_all_built_testcases()

    def get_built_testcases(self) -> List[Tuple[TestcaseGroup, List[Testcase]]]:
        pkg = package.find_problem_package_or_die()
        tests_per_group = self.get_built_testcases_per_group()
        return [(group, tests_per_group[group.name]) for group in pkg.testcases]

    def get_flattened_built_testcases(self) -> List[Testcase]:
        pkg = package.find_problem_package_or_die()
        tests_per_group = self.get_built_testcases_per_group()

        res = []
        for group in pkg.testcases:
            res.extend(tests_per_group[group.name])
        return res

    def get_statement_for_language(self, lang: str) -> Statement:
        pkg = package.find_problem_package_or_die()
        for statement in pkg.expanded_statements:
            if statement.language == lang:
                return statement
        raise


class BaseContestPackager(ABC):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass

    @abstractmethod
    def package(
        self,
        built_packages: List[BuiltProblemPackage],
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltContestStatement],
    ) -> pathlib.Path:
        pass

    def languages(self):
        pkg = contest_package.find_contest_package_or_die()

        res = set()
        for statement in pkg.expanded_statements:
            res.add(statement.language)
        return list(res)

    def statement_types(self) -> List[StatementType]:
        return [StatementType.PDF]

    def get_statement_for_language(self, lang: str) -> ContestStatement:
        contest = contest_package.find_contest_package_or_die()
        for statement in contest.expanded_statements:
            if statement.language == lang:
                return statement
        raise


class ContestZipper(BaseContestPackager):
    def __init__(
        self, filename: str, zip_inner: bool = False, prefer_shortname: bool = True
    ):
        super().__init__()
        self.zip_inner = zip_inner
        self.filename = filename
        self.prefer_shortname = prefer_shortname

    def package(
        self,
        built_packages: List[BuiltProblemPackage],
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltContestStatement],
    ) -> pathlib.Path:
        for built_package in built_packages:
            if self.prefer_shortname:
                pkg_path = into_path / 'problems' / built_package.problem.short_name
            else:
                pkg_path = into_path / 'problems' / built_package.package.name

            if self.zip_inner:
                pkg_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(built_package.path, pkg_path.with_suffix('.zip'))
            else:
                pkg_path.mkdir(parents=True, exist_ok=True)
                shutil.unpack_archive(built_package.path, pkg_path, format='zip')

        # Zip all.
        shutil.make_archive(str(build_path / self.filename), 'zip', into_path)

        return build_path / pathlib.Path(self.filename).with_suffix('.zip')


async def run_packager(
    packager_cls: Type[BasePackager],
    verification: environment.VerificationParam,
    **kwargs,
) -> pathlib.Path:
    from rbx.box import builder

    header.generate_header()

    if limits_info.get_saved_limits_profile(packager_cls.name()) is not None:
        console.console.print(
            f'[warning]Using saved limits profile for [item]{packager_cls.name()}[/item].[/warning]'
        )

    with limits_info.use_profile(packager_cls.name()):
        if not await builder.verify(verification=verification):
            console.console.print(
                '[error]Build or verification failed, check the report.[/error]'
            )
            raise typer.Exit(1)

    pkg = package.find_problem_package_or_die()

    if pkg.type not in packager_cls.task_types():
        console.console.print(
            f'[error]Packager [item]{packager_cls.name()}[/item] does not support task type [item]{pkg.type}[/item].[/error]'
        )
        raise typer.Exit(1)

    packager = packager_cls(**kwargs)

    statement_types = packager.statement_types()
    built_statements = []

    with limits_info.use_profile(packager_cls.name()):
        for statement_type in statement_types:
            languages = packager.languages()
            for language in languages:
                statement = packager.get_statement_for_language(language)
                statement_path = build_statement(statement, pkg, statement_type)
                built_statements.append(
                    BuiltStatement(statement, statement_path, statement_type)
                )

    console.console.print(f'Packaging problem for [item]{packager.name()}[/item]...')

    with tempfile.TemporaryDirectory() as td, limits_info.use_profile(
        packager_cls.name()
    ):
        result_path = packager.package(
            package.get_build_path(), pathlib.Path(td), built_statements
        )

    console.console.print(
        f'[success]Problem packaged for [item]{packager.name()}[/item]![/success]'
    )
    console.console.print(f'Package was saved at {href(result_path)}')
    return result_path



import pathlib
from typing import Iterable, List, Tuple

from rbx import console
from rbx.box.cd import (
    find_all_ancestor_packages,
    is_contest_package,
    is_problem_package,
)
from rbx.box.contest.contest_package import find_contest, find_contest_package_or_die
from rbx.box.formatting import get_formatted_memory


def find_problem_packages_from_contest(
    root: pathlib.Path = pathlib.Path(),
) -> Iterable[pathlib.Path]:
    contest_path = find_contest(root)
    contest = find_contest_package_or_die(contest_path)
    for problem in contest.problems:
        yield contest_path / problem.get_path()


def find_all_reachable_packages(
    root: pathlib.Path = pathlib.Path(),
) -> List[pathlib.Path]:
    packages = find_all_ancestor_packages(root)

    for package in list(packages):
        if is_contest_package(package):
            packages.extend(find_problem_packages_from_contest(package))
    return packages


def find_and_group_all_reachable_packages(
    root: pathlib.Path = pathlib.Path(),
) -> Tuple[List[pathlib.Path], List[pathlib.Path]]:
    packages = find_all_reachable_packages(root)
    contest_packages = set(pkg for pkg in packages if is_contest_package(pkg))
    problem_packages = set(pkg for pkg in packages if is_problem_package(pkg))
    return sorted(contest_packages), sorted(problem_packages)


def get_dir_size(path: pathlib.Path) -> int:
    if not path.is_dir():
        return 0
    return sum(
        f.stat().st_size
        for f in path.glob('**/*')
        if f.is_file() and not f.is_symlink()
    )


def get_cache_size(root: pathlib.Path = pathlib.Path()) -> int:
    cache_dir = root / '.box'
    return get_dir_size(cache_dir)


def get_build_size(root: pathlib.Path = pathlib.Path()) -> int:
    build_dir = root / 'build'
    return get_dir_size(build_dir)


def print_package_stats(root: pathlib.Path = pathlib.Path()) -> int:
    if is_contest_package(root):
        console.console.print(f'[status]Contest package[/status]: [item]{root}[/item]')
    else:
        console.console.print(f'[status]Problem package[/status]: [item]{root}[/item]')

    cache_size = get_cache_size(root)
    build_size = get_build_size(root)
    console.console.print(
        f'[status]Cache size[/status]: [item]{get_formatted_memory(cache_size)}[/item]'
    )
    console.console.print(
        f'[status]Build size[/status]: [item]{get_formatted_memory(build_size)}[/item]'
    )

    return cache_size + build_size


def print_global_stats() -> int:
    cache_size = get_cache_size()
    console.console.print(
        f'[status]Global cache size[/status]: [item]{get_formatted_memory(cache_size)}[/item]'
    )
    return cache_size


def print_reachable_package_stats(root: pathlib.Path = pathlib.Path()) -> None:
    contest_packages, problem_packages = find_and_group_all_reachable_packages(root)
    total_size = 0
    for pkg in contest_packages:
        total_size += print_package_stats(pkg)
        console.console.print()
    for pkg in problem_packages:
        total_size += print_package_stats(pkg)
        console.console.print()

    total_size += print_global_stats()
    console.console.print(
        f'[status]Total size[/status]: [item]{get_formatted_memory(total_size)}[/item]'
    )



import asyncio
import pathlib
from typing import Optional

import rich.text
from textual import work
from textual.reactive import reactive

from rbx.box import testcase_utils
from rbx.box.ui.widgets.rich_log_box import RichLogBox

BATCH_SIZE = 1024


class InteractionBox(RichLogBox):
    DEFAULT_CSS = """
    InteractionBox {
        border: solid $accent;
        height: 1fr;
        width: 1fr;
    }
    """

    path: reactive[Optional[pathlib.Path]] = reactive(None)

    def on_mount(self):
        super().on_mount()
        self.auto_scroll = False
        self.can_focus = False

    def _show_raw_text(self, path: pathlib.Path):
        path_str = str(path.relative_to(pathlib.Path.cwd()))
        self.write(
            rich.text.Text(
                'Showing raw interaction file because the interaction text is not parseable.\n'
                'This might usually happen when the processes do not communicate properly.\n',
                style='red',
            )
        )
        self.write(rich.text.Text(path.read_text()))
        self.border_subtitle = f'{path_str} (raw file)'

    @work(exclusive=True)
    async def _load_file(self, path: pathlib.Path):
        self.clear()
        path_str = str(path.relative_to(pathlib.Path.cwd()))
        self.border_subtitle = f'{path_str} (loading...)'

        try:
            interaction = await asyncio.to_thread(
                testcase_utils.parse_interaction, path
            )
        except testcase_utils.TestcaseInteractionParsingError:
            self._show_raw_text(path)
            return

        for entry in interaction.entries:
            if entry.pipe == 0:
                self.write(rich.text.Text(entry.data.rstrip(), style='green'))
            else:
                self.write(rich.text.Text(entry.data.rstrip()))

        self.border_subtitle = path_str

    async def watch_path(self, path: Optional[pathlib.Path]):
        self.clear()

        if path is None:
            self.border_subtitle = '(no file selected)'
            return

        if not path.is_file():
            path_str = str(path.relative_to(pathlib.Path.cwd()))
            self.border_subtitle = f'{path_str} (does not exist)'
            return

        self._load_file(path)



import pathlib
import shutil
from typing import List, Optional

import typer

from rbx import console, utils
from rbx.box import header, limits_info, package
from rbx.box.lang import code_to_lang, code_to_langs, is_valid_lang_code
from rbx.box.packaging.packager import (
    BaseContestPackager,
    BasePackager,
    BuiltContestStatement,
    BuiltProblemPackage,
    BuiltStatement,
)
from rbx.box.packaging.polygon import xml_schema as polygon_schema
from rbx.box.schema import TaskType
from rbx.config import get_testlib

DAT_TEMPLATE = """
@contest "{name}"
@contlen 300
@problems {problems}
@teams 0
@submissions 0
"""


def _select_main_language(
    names: List[polygon_schema.Name], main_language: Optional[str]
):
    if names:
        if main_language is not None:
            lang_name = code_to_lang(main_language)
            found = False
            for name in names:
                if name.language == lang_name:
                    name.main = True
                    found = True
                    break
            if not found:
                console.console.print(
                    f'[error]Main language [item]{main_language}[/item] not found.[/error]'
                )
                raise typer.Exit(1)
        else:
            names[0].main = True


class PolygonPackager(BasePackager):
    def __init__(self, main_language: Optional[str] = None):
        self.main_language = main_language

    @classmethod
    def task_types(cls) -> List[TaskType]:
        return [TaskType.BATCH, TaskType.COMMUNICATION]

    def _validate(self):
        langs = self.languages()
        pkg = package.find_problem_package_or_die()

        lang_codes = set()
        for statement in pkg.expanded_statements:
            lang_codes.add(statement.title)

        for lang in langs:
            if lang not in lang_codes:
                console.console.print(
                    f'[error]No statement from language [item]{lang}[/item] found. '
                    'Polygon needs one statement for each language.[/error]'
                )
                raise typer.Exit(1)

    def _get_names(self) -> List[polygon_schema.Name]:
        names = [
            polygon_schema.Name(
                language=code_to_langs([lang])[0],
                value=self.get_statement_for_language(lang).title,
            )
            for lang in self.languages()
        ]

        _select_main_language(names, self.main_language)

        return names

    def _get_checker(self) -> polygon_schema.Checker:
        # TODO: support other checker languages
        return polygon_schema.Checker(
            name='rbx::checker',
            type='testlib',
            source=polygon_schema.File(path='files/check.cpp', type='cpp.g++17'),
            cpy=polygon_schema.File(path='check.cpp'),
        )

    def _get_interactor(self) -> Optional[polygon_schema.Interactor]:
        pkg = package.find_problem_package_or_die()
        if pkg.interactor is None:
            return None
        return polygon_schema.Interactor(
            source=polygon_schema.File(path='files/interactor.cpp', type='cpp.g++17'),
        )

    def _get_manual_test(self) -> polygon_schema.Test:
        # TODO: return samples
        return polygon_schema.Test(method='manual')

    def _get_single_testset(self) -> polygon_schema.Testset:
        pkg = package.find_problem_package_or_die()

        if pkg.modifiers:
            console.console.print(
                f'[warning]Your package specify custom limit modifiers for languages {list(pkg.modifiers)}.[/warning]'
            )
            console.console.print(
                '[warning]Polygon packages do not support language-based limit modifiers, so they will not be applied to the supplied limits.[/warning]'
            )

        console.console.print(
            f'[warning]Polygon packages cannot honor the {pkg.outputLimit}kb output limit.'
        )

        testcases = self.get_flattened_built_testcases()

        limits = limits_info.get_limits(None, profile='polygon')
        assert limits.time is not None
        assert limits.memory is not None

        return polygon_schema.Testset(
            name='tests',
            timelimit=limits.time,
            memorylimit=limits.memory * 1024 * 1024,
            size=len(testcases),
            inputPattern='tests/%03d',
            answerPattern='tests/%03d.a',
            tests=[self._get_manual_test() for _ in range(len(testcases))],
        )

    def _get_judging(self) -> polygon_schema.Judging:
        return polygon_schema.Judging(testsets=[self._get_single_testset()])

    def _get_files(self) -> List[polygon_schema.File]:
        return [
            polygon_schema.File(path='files/testlib.h', type='h.g++'),
            polygon_schema.File(path='files/rbx.h', type='h.g++'),
        ]

    def _statement_application_type(self, statement: BuiltStatement) -> str:
        return 'application/pdf'

    def _process_statement(
        self,
        built_statement: BuiltStatement,
        into_path: pathlib.Path,
    ) -> polygon_schema.Statement:
        language = code_to_langs([built_statement.statement.language])[0]
        final_path = into_path / 'statements' / language / built_statement.path.name

        final_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(built_statement.path, final_path)

        return polygon_schema.Statement(
            path=str(utils.abspath(final_path).relative_to(utils.abspath(into_path))),
            language=language,
            type=self._statement_application_type(built_statement),  # type: ignore
        )

    def _process_statements(
        self, built_statements: List[BuiltStatement], into_path: pathlib.Path
    ) -> List[polygon_schema.Statement]:
        return [
            self._process_statement(built_statement, into_path)
            for built_statement in built_statements
        ]

    @classmethod
    def name(cls) -> str:
        return 'polygon'

    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ):
        pkg = package.find_problem_package_or_die()

        problem = polygon_schema.Problem(
            short_name=pkg.name,
            names=self._get_names(),
            checker=self._get_checker(),
            judging=self._get_judging(),
            files=self._get_files(),
            interactor=self._get_interactor(),
            # TODO: revisit polygon problem statements
            # statements=self._process_statements(built_statements, into_path),
        )

        descriptor: str = problem.to_xml(
            skip_empty=True,
            encoding='utf-8',
            pretty_print=True,
            standalone=True,
        )  # type: ignore
        if isinstance(descriptor, bytes):
            descriptor = descriptor.decode()

        # Prepare files
        files_path = into_path / 'files'
        files_path.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(get_testlib(), files_path / 'testlib.h')
        shutil.copyfile(header.get_header(), files_path / 'rbx.h')
        shutil.copyfile(package.get_checker().path, files_path / 'check.cpp')
        shutil.copyfile(package.get_checker().path, into_path / 'check.cpp')
        if pkg.interactor is not None:
            shutil.copyfile(pkg.interactor.path, files_path / 'interactor.cpp')

        # Copy all testcases
        (into_path / 'tests').mkdir(parents=True, exist_ok=True)
        testcases = self.get_flattened_built_testcases()
        for i, testcase in enumerate(testcases):
            shutil.copyfile(
                testcase.inputPath,
                into_path / f'tests/{i + 1:03d}',
            )
            if testcase.outputPath is not None:
                shutil.copyfile(
                    testcase.outputPath,
                    into_path / f'tests/{i + 1:03d}.a',
                )
            else:
                (into_path / f'tests/{i + 1:03d}.a').touch()

        # Write problem.xml
        (into_path / 'problem.xml').write_text(descriptor)

        # Zip all.
        shutil.make_archive(str(build_path / 'problem'), 'zip', into_path)

        return build_path / 'problem.zip'


class PolygonContestPackager(BaseContestPackager):
    def __init__(self, main_language: Optional[str] = None):
        self.main_language = main_language

    @classmethod
    def name(cls) -> str:
        return 'polygon'

    def _get_names(self) -> List[polygon_schema.Name]:
        names = [
            polygon_schema.Name(
                language=code_to_langs([lang])[0],
                value=self.get_statement_for_language(lang).title,
            )
            for lang in self.languages()
            if is_valid_lang_code(lang)
        ]
        _select_main_language(names, self.main_language)
        return names

    def _process_statement(
        self,
        built_statement: BuiltContestStatement,
        into_path: pathlib.Path,
    ) -> polygon_schema.Statement:
        language = code_to_langs([built_statement.statement.language])[0]
        final_path = into_path / 'statements' / language / built_statement.path.name

        final_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(built_statement.path, final_path)

        return polygon_schema.Statement(
            path=str(utils.abspath(final_path).relative_to(utils.abspath(into_path))),
            language=language,
            type='application/pdf',  # type: ignore
        )

    def _process_statements(
        self, built_statements: List[BuiltContestStatement], into_path: pathlib.Path
    ) -> List[polygon_schema.Statement]:
        return [
            self._process_statement(built_statement, into_path)
            for built_statement in built_statements
        ]

    def _get_problems(
        self, built_packages: List[BuiltProblemPackage]
    ) -> List[polygon_schema.ContestProblem]:
        return [
            polygon_schema.ContestProblem(
                index=built_package.problem.short_name,
                path=str(pathlib.Path('problems') / built_package.problem.short_name),
            )
            for built_package in built_packages
        ]

    def _get_dat(self, built_packages: List[BuiltProblemPackage]) -> str:
        dat = DAT_TEMPLATE.format(
            name=self.get_statement_for_language(self.languages()[0]).title,
            problems=len(built_packages),
        )
        return (
            dat
            + '\n'.join(
                f'@p {pkg.problem.short_name},{pkg.problem.short_name},20,0'
                for pkg in built_packages
            )
            + '\n'
        )

    def package(
        self,
        built_packages: List[BuiltProblemPackage],
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltContestStatement],
    ) -> pathlib.Path:
        filtered_statements = []
        for built_statement in built_statements:
            if not is_valid_lang_code(built_statement.statement.language):
                continue
            filtered_statements.append(built_statement)

        for built_package in built_packages:
            pkg_path = into_path / 'problems' / built_package.problem.short_name
            pkg_path.mkdir(parents=True, exist_ok=True)

            shutil.unpack_archive(built_package.path, pkg_path, format='zip')

        # Build contest descriptor.
        contest = polygon_schema.Contest(
            names=self._get_names(),
            statements=self._process_statements(filtered_statements, into_path),
            problems=self._get_problems(built_packages),
        )
        descriptor: str = contest.to_xml(
            skip_empty=True,
            encoding='utf-8',
            pretty_print=True,
            standalone=True,
        )  # type: ignore
        if isinstance(descriptor, bytes):
            descriptor = descriptor.decode()

        # Write contest.xml
        (into_path / 'contest.xml').write_text(descriptor)

        # Write contest.dat
        (into_path / 'contest.dat').write_text(self._get_dat(built_packages))

        # Zip all.
        shutil.make_archive(str(build_path / 'contest'), 'zip', into_path)

        return build_path / 'contest.zip'



import contextvars
import dataclasses
from dataclasses import dataclass


@dataclass(frozen=True)
class DebugContext:
    enable: bool = False


debug_var = contextvars.ContextVar('debug', default=DebugContext())


def get_debug_context() -> DebugContext:
    return debug_var.get()


class Debug:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.token = None

    def __enter__(self):
        self.token = debug_var.set(
            dataclasses.replace(debug_var.get(), *self.args, **self.kwargs)
        )

    def __exit__(self, exc_type, exc_value, traceback):
        if self.token is not None:
            debug_var.reset(self.token)



import pathlib
from concurrent.futures import ThreadPoolExecutor

from rich.progress import MofNCompleteColumn, Progress, SpinnerColumn

from rbx.box.tooling.boca.scraper import BocaRun, BocaScraper


def scrape_boca(into_path: pathlib.Path):
    scraper = BocaScraper()
    scraper.login()
    runs = scraper.list_runs()

    progress = Progress(
        SpinnerColumn(),
        *Progress.get_default_columns(),
        MofNCompleteColumn(),
        transient=True,
    )
    scrape_task = progress.add_task('Scraping runs...', total=len(runs))
    with progress:

        def work(run: BocaRun):
            scraper.download_run(
                run.run_number,
                run.site_number,
                pathlib.Path(into_path) / run.problem_shortname,
                name=f'{run.run_number}-{run.site_number}-{run.outcome.short_name().lower()}',
            )

            progress.update(scrape_task, advance=1)

        with ThreadPoolExecutor(max_workers=10) as executor:
            executor.map(work, runs)



import dataclasses
import pathlib
from typing import Optional

from textual.app import ComposeResult
from textual.containers import Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import ContentSwitcher

from rbx.box.ui.widgets.file_log import FileLog
from rbx.box.ui.widgets.interaction_box import InteractionBox
from rbx.box.ui.widgets.rich_log_box import RichLogBox


@dataclasses.dataclass
class TestcaseRenderingData:
    input_path: Optional[pathlib.Path] = None
    output_path: Optional[pathlib.Path] = None
    stderr_path: Optional[pathlib.Path] = None
    interaction_path: Optional[pathlib.Path] = None
    log_path: Optional[pathlib.Path] = None
    rich_content: Optional[str] = None

    @classmethod
    def from_one_path(cls, path: pathlib.Path) -> 'TestcaseRenderingData':
        return cls(
            input_path=path.with_suffix('.in'),
            output_path=path.with_suffix('.out'),
            stderr_path=path.with_suffix('.err'),
            log_path=path.with_suffix('.log'),
            interaction_path=path.with_suffix('.pio'),
        )


class TestBoxWidget(Widget, can_focus=False):
    data: reactive[TestcaseRenderingData] = reactive(
        default=lambda: TestcaseRenderingData(),
        bindings=True,
    )

    @dataclasses.dataclass
    class Logs:
        output: FileLog
        stderr: FileLog
        log: FileLog
        interaction: InteractionBox

    def logs(self) -> Logs:
        return self.Logs(
            output=self.query_one('#test-box-output', FileLog),
            stderr=self.query_one('#test-box-stderr', FileLog),
            log=self.query_one('#test-box-log', FileLog),
            interaction=self.query_one('#test-box-interaction', InteractionBox),
        )

    def compose(self) -> ComposeResult:
        with Vertical():
            with ContentSwitcher(initial='test-box-output', id='test-box-switcher'):
                yield FileLog(id='test-box-output')
                yield FileLog(id='test-box-stderr')
                yield FileLog(id='test-box-log')
                yield InteractionBox(id='test-box-interaction')
            yield RichLogBox(id='test-box-metadata')

    def on_mount(self):
        logs = self.logs()
        logs.output.border_title = 'Output'
        logs.stderr.border_title = 'Stderr'
        logs.log.border_title = 'Log'
        logs.interaction.border_title = 'Interaction'
        metadata = self.query_one('#test-box-metadata', RichLogBox)
        metadata.display = False
        metadata.border_title = 'Metadata'
        metadata.wrap = True
        metadata.markup = True
        metadata.clear()
        metadata.write('No metadata')

        self.watch_data(self.data)

    def watch_data(self, data: TestcaseRenderingData):
        logs = self.logs()
        logs.output.path = data.output_path
        logs.stderr.path = data.stderr_path
        logs.log.path = data.log_path
        logs.interaction.path = data.interaction_path
        metadata = self.query_one('#test-box-metadata', RichLogBox)
        metadata.clear()
        if data.rich_content is not None:
            metadata.write(data.rich_content)
        else:
            metadata.write('No metadata')

    def reset(self):
        self.data = TestcaseRenderingData()

    def show_output(self):
        self.query_one(ContentSwitcher).current = 'test-box-output'

    def show_stderr(self):
        self.query_one(ContentSwitcher).current = 'test-box-stderr'

    def show_log(self):
        self.query_one(ContentSwitcher).current = 'test-box-log'

    def show_interaction(self):
        self.query_one(ContentSwitcher).current = 'test-box-interaction'

    def toggle_metadata(self):
        metadata = self.query_one('#test-box-metadata', RichLogBox)
        metadata.display = not metadata.display
        metadata.refresh()



from typing import Optional, Set

from rbx import console, utils
from rbx.box import environment, package
from rbx.box.environment import VerificationLevel
from rbx.box.generators import (
    generate_outputs_for_testcases,
    generate_testcases,
)
from rbx.box.solutions import (
    is_fast,
    print_run_report,
    run_solutions,
)
from rbx.box.testcase_extractors import extract_generation_testcases_from_groups
from rbx.box.validators import (
    has_validation_errors,
    print_validation_report,
    validate_testcases,
)


async def build(
    verification: environment.VerificationParam,
    groups: Optional[Set[str]] = None,
    output: Optional[bool] = True,
) -> bool:
    no_main_solution_report = False
    if output is None:
        output = package.get_main_solution() is not None
        no_main_solution_report = not output

    with utils.StatusProgress(
        'Building testcases...',
        'Built [item]{processed}[/item] testcases...',
        keep=True,
    ) as s:
        await generate_testcases(s, groups=groups)

    if verification > 0:
        validator = package.get_validator_or_nil()
        if validator is None:
            console.console.print(
                '[warning]No validator found, skipping validation.[/warning]'
            )

        if validator is not None:
            with utils.StatusProgress(
                'Validating testcases...',
                'Validated [item]{processed}[/item] testcases...',
                keep=True,
            ) as s:
                infos = await validate_testcases(
                    s,
                    groups=groups,
                )
                print_validation_report(infos)

            if has_validation_errors(infos):
                console.console.print(
                    '[error]Validation failed, check the report above.[/error]'
                )
                return False

    with utils.StatusProgress(
        'Building outputs for testcases...',
        'Built [item]{processed}[/item] outputs...',
        keep=True,
    ) as s:
        if output:
            entries = [
                entry.group_entry
                for entry in await extract_generation_testcases_from_groups(groups)
            ]
            await generate_outputs_for_testcases(entries, s)

    console.console.print(
        '[success]Problem built.[/success] '
        '[warning]Check the output for verification errors![/warning]'
    )

    if no_main_solution_report:
        console.console.print(
            '[warning]No main solution found, skipping generating samples for the statement.[/warning]'
        )

    return True


async def verify(verification: environment.VerificationParam) -> bool:
    if not await build(verification=verification):
        return False

    if verification < VerificationLevel.FAST_SOLUTIONS.value:
        return True

    tracked_solutions = None
    if verification < VerificationLevel.ALL_SOLUTIONS.value:
        tracked_solutions = {
            str(solution.path)
            for solution in package.get_solutions()
            if is_fast(solution)
        }

    with utils.StatusProgress('Running solutions...') as s:
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            verification=VerificationLevel(verification),
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    return await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
    )



import pathlib
import tempfile
import typing
from typing import Annotated, Any, Dict, List, Optional, Tuple

import syncer
import typer

from rbx import annotations, console, utils
from rbx.box import environment, limits_info, naming, package
from rbx.box.formatting import href
from rbx.box.schema import Package, expand_any_vars
from rbx.box.statements.builders import (
    BUILDER_LIST,
    PROBLEM_BUILDER_LIST,
    StatementBuilder,
    StatementBuilderContext,
    StatementBuilderProblem,
    StatementCodeLanguage,
    StatementSample,
    prepare_assets,
)
from rbx.box.statements.schema import (
    ConversionStep,
    ConversionType,
    Statement,
    StatementType,
)
from rbx.box.testcase_utils import get_samples

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


def get_environment_languages_for_statement() -> List[StatementCodeLanguage]:
    env = environment.get_environment()

    res = []
    for language in env.languages:
        cmd = ''
        compilation_cfg = environment.get_compilation_config(language.name)
        cmd = ' && '.join(compilation_cfg.commands or [])
        if not cmd:
            execution_cfg = environment.get_execution_config(language.name)
            cmd = execution_cfg.command

        res.append(
            StatementCodeLanguage(
                id=language.name,
                name=language.readableName or language.name,
                command=cmd or '',
            )
        )

    return res


def get_builder(
    name: ConversionType, builder_list: List[StatementBuilder]
) -> StatementBuilder:
    candidates = [builder for builder in builder_list if builder.name() == name]
    if not candidates:
        console.console.print(
            f'[error]No statement builder found with name [name]{name}[/name][/error]'
        )
        raise typer.Exit(1)
    return candidates[0]


def get_implicit_builders(
    input_type: StatementType, output_type: StatementType
) -> Optional[List[StatementBuilder]]:
    par: Dict[StatementType, Optional[StatementBuilder]] = {input_type: None}

    def _iterate() -> bool:
        nonlocal par
        for bdr in BUILDER_LIST:
            u = bdr.input_type()
            if u not in par:
                continue
            v = bdr.output_type()
            if v in par:
                continue
            par[v] = bdr
            return True
        return False

    while _iterate() and output_type not in par:
        pass

    if output_type not in par:
        return None

    res = []
    cur = output_type
    while par[cur] is not None:
        res.append(par[cur])
        cur = typing.cast(StatementBuilder, par[cur]).input_type()

    return list(reversed(res))


def _try_implicit_builders(
    statement_id: str, input_type: StatementType, output_type: StatementType
) -> List[StatementBuilder]:
    implicit_builders = get_implicit_builders(input_type, output_type)
    if implicit_builders is None:
        console.console.print(
            f'[error]Cannot implicitly convert statement [item]{statement_id}[/item] '
            f'from [item]{input_type}[/item] '
            f'to specified output type [item]{output_type}[/item].[/error]'
        )
        raise typer.Exit(1)
    console.console.print(
        'Implicitly adding statement builders to convert statement '
        f'from [item]{input_type}[/item] to [item]{output_type}[/item]...'
    )
    return implicit_builders


def _get_configured_params_for(
    configure: List[ConversionStep], conversion_type: ConversionType
) -> Optional[ConversionStep]:
    for step in configure:
        if step.type == conversion_type:
            return step
    return None


def get_builders(
    statement_id: str,
    steps: List[ConversionStep],
    configure: List[ConversionStep],
    input_type: StatementType,
    output_type: Optional[StatementType],
    builder_list: List[StatementBuilder] = BUILDER_LIST,
) -> List[Tuple[StatementBuilder, ConversionStep]]:
    last_output = input_type
    builders: List[Tuple[StatementBuilder, ConversionStep]] = []

    # Conversion steps to force during build.
    for step in steps:
        builder = get_builder(step.type, builder_list=builder_list)
        if builder.input_type() != last_output:
            implicit_builders = _try_implicit_builders(
                statement_id, last_output, builder.input_type()
            )
            builders.extend(
                (builder, builder.default_params()) for builder in implicit_builders
            )
        builders.append((builder, step))
        last_output = builder.output_type()

    if output_type is not None and last_output != output_type:
        implicit_builders = _try_implicit_builders(
            statement_id, last_output, output_type
        )
        builders.extend(
            (builder, builder.default_params()) for builder in implicit_builders
        )

    # Override statement configs.
    def reconfigure(params: ConversionStep) -> ConversionStep:
        new_params = _get_configured_params_for(configure, params.type)
        return new_params or params

    reconfigured_builders = [
        (builder, reconfigure(params)) for builder, params in builders
    ]
    return reconfigured_builders


def get_relative_assets(
    relative_to: pathlib.Path,
    assets: List[str],
) -> List[Tuple[pathlib.Path, pathlib.Path]]:
    relative_to = utils.abspath(relative_to)
    if not relative_to.is_dir():
        relative_to = relative_to.parent
    res = []
    for asset in assets:
        relative_path = pathlib.Path(asset)
        if not relative_path.is_file():
            globbed = list(
                path
                for path in pathlib.Path().glob(str(relative_path))
                if path.is_file()
            )
            if not globbed and '*' not in str(relative_path):
                console.console.print(
                    f'[error]Asset [item]{asset}[/item] does not exist.[/error]'
                )
                raise typer.Exit(1)
            res.extend(get_relative_assets(relative_to, list(map(str, globbed))))
            continue
        if not utils.abspath(relative_path).is_relative_to(relative_to):
            console.console.print(
                f'[error]Asset [item]{asset}[/item] is not relative to your statement.[/error]'
            )
            raise typer.Exit(1)

        res.append(
            (
                utils.abspath(relative_path),
                utils.abspath(relative_path).relative_to(relative_to),
            )
        )

    return res


def build_statement_bytes(
    statement: Statement,
    pkg: Package,
    output_type: Optional[StatementType] = None,
    short_name: Optional[str] = None,
    overridden_params_root: pathlib.Path = pathlib.Path(),
    overridden_params: Optional[Dict[ConversionType, ConversionStep]] = None,
    overridden_assets: Optional[List[Tuple[pathlib.Path, pathlib.Path]]] = None,
    use_samples: bool = True,
    custom_vars: Optional[Dict[str, Any]] = None,
) -> Tuple[bytes, StatementType]:
    overridden_params = overridden_params or {}
    overridden_assets = overridden_assets or []

    if not statement.path.is_file():
        console.console.print(
            f'[error]Statement file [item]{statement.path}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)
    builders = get_builders(
        str(statement.path),
        statement.steps,
        statement.configure,
        statement.type,
        output_type,
        builder_list=PROBLEM_BUILDER_LIST,
    )
    last_output = statement.type
    last_content = statement.path.read_bytes()
    for bdr, params in builders:
        with tempfile.TemporaryDirectory() as td:
            # Here, create a new temp context for each builder call.
            assets = get_relative_assets(statement.path, statement.assets)

            # Use either overridden assets (by contest) or usual assets.
            # Remember to modify the root to contest root if that's the case.
            if bdr.name() in overridden_params:
                assets.extend(
                    bdr.inject_assets(
                        overridden_params_root, overridden_params[bdr.name()]
                    )
                )
            else:
                assets.extend(bdr.inject_assets(pathlib.Path(), params))
            assets.extend(overridden_assets)

            prepare_assets(assets, pathlib.Path(td))
            output = bdr.build(
                input=last_content,
                context=StatementBuilderContext(
                    lang=statement.language,
                    languages=get_environment_languages_for_statement(),
                    params=params,
                    root=pathlib.Path(td),
                ),
                item=StatementBuilderProblem(
                    limits=limits_info.get_limits_profile(
                        profile=limits_info.get_active_profile()
                    ),
                    package=pkg,
                    statement=statement,
                    samples=StatementSample.from_testcases(
                        get_samples() if use_samples else []
                    ),
                    short_name=short_name,
                    vars={
                        **pkg.expanded_vars,
                        **statement.expanded_vars,
                        **(custom_vars or {}),
                    },
                ),
                verbose=False,
            )
        last_output = bdr.output_type()
        last_content = output

    return last_content, last_output


def build_statement(
    statement: Statement,
    pkg: Package,
    output_type: Optional[StatementType] = None,
    use_samples: bool = True,
    custom_vars: Optional[Dict[str, Any]] = None,
) -> pathlib.Path:
    last_content, last_output = build_statement_bytes(
        statement,
        pkg,
        output_type=output_type,
        use_samples=use_samples,
        custom_vars=custom_vars,
        short_name=naming.get_problem_shortname(),
    )
    statement_path = (package.get_build_path() / statement.name).with_suffix(
        last_output.get_file_suffix()
    )
    active_profile = limits_info.get_active_profile()
    if (
        active_profile is not None
        and limits_info.get_saved_limits_profile(active_profile) is not None
    ):
        statement_path = statement_path.with_stem(
            f'{statement_path.stem}-{active_profile}'
        )
    statement_path.parent.mkdir(parents=True, exist_ok=True)
    statement_path.write_bytes(last_content)
    console.console.print(
        f'Statement [item]{statement.name}[/item] built successfully for language '
        f'[item]{statement.language}[/item] at '
        f'{href(statement_path)}'
    )
    return statement_path


@app.command('build, b', help='Build statements.')
@package.within_problem
@syncer.sync
async def build(
    verification: environment.VerificationParam,
    names: Annotated[
        Optional[List[str]],
        typer.Argument(
            help='Names of statements to build.',
        ),
    ] = None,
    languages: Annotated[
        Optional[List[str]],
        typer.Option(
            help='Languages to build statements for. If not specified, build statements for all available languages.',
        ),
    ] = None,
    output: Annotated[
        Optional[StatementType],
        typer.Option(
            case_sensitive=False,
            help='Output type to be generated. If not specified, will infer from the conversion steps specified in the package.',
        ),
    ] = StatementType.PDF,
    samples: Annotated[
        bool,
        typer.Option(help='Whether to build the statement with samples or not.'),
    ] = True,
    vars: Annotated[
        Optional[List[str]],
        typer.Option(
            '--vars',
            help='Variables to be used in the statements.',
        ),
    ] = None,
):
    # At most run the validators, only in samples.
    if samples:
        from rbx.box import builder

        if not await builder.build(
            verification=verification,
            groups=set(['samples']),
            output=None,
        ):
            console.console.print(
                '[error]Failed to build statements with samples, aborting.[/error]'
            )
            raise typer.Exit(1)

    pkg = package.find_problem_package_or_die()
    candidate_languages = set(languages or [])
    candidate_names = set(names or [])

    def should_process(st: Statement) -> bool:
        if candidate_languages and st.language not in candidate_languages:
            return False
        if candidate_names and st.name not in candidate_names:
            return False
        return True

    valid_statements = [st for st in pkg.expanded_statements if should_process(st)]

    if not valid_statements:
        console.console.print(
            '[error]No statement found according to the specified criteria.[/error]',
        )
        raise typer.Exit(1)

    for statement in valid_statements:
        build_statement(
            statement,
            pkg,
            output_type=output,
            use_samples=samples,
            custom_vars=expand_any_vars(annotations.parse_dictionary_items(vars)),
        )


@app.callback()
def callback():
    pass



from __future__ import annotations

import pathlib
from enum import Enum
from typing import Annotated, List, Literal, Optional, Union

from pydantic import AfterValidator, BaseModel, ConfigDict, Field

from rbx.autoenum import AutoEnum, alias
from rbx.box.fields import FNameField, RecVars, Vars, expand_vars
from rbx.box.lang import is_valid_lang_code


def validate_statement_language(lang: str):
    if not is_valid_lang_code(lang) or not lang.islower():
        raise ValueError(
            f'Invalid statement language: {lang}. Language must be a valid lowercase ISO 639-1 code.'
        )
    return lang


StatementLanguage = Annotated[str, AfterValidator(validate_statement_language)]


### Conversion types
class ConversionType(str, Enum):
    rbxToTex = 'rbx-tex'
    """Conversion from rbxTeX to LaTeX."""

    rbxMarkdownToTeX = 'rbx-md-tex'
    """Conversion from rbxMarkdown to LaTeX."""
    TexToPDF = 'tex2pdf'
    """Conversion from LaTeX to PDF using pdfLaTeX."""

    JinjaTeX = 'jinja-tex'
    """Conversion from LaTeX with Jinja2 expressions to LaTeX."""

    def __repr__(self):
        return str.__repr__(self.value)


### Conversion nodes.
class rbxMarkdownToTeX(BaseModel):
    """Configures the conversion between rbxMarkdown and LaTeX."""

    type: Literal[ConversionType.rbxMarkdownToTeX]


class rbxToTeX(BaseModel):
    """Configures the conversion between rbxTeX and LaTeX."""

    type: Literal[ConversionType.rbxToTex]

    template: pathlib.Path = Field(
        default=pathlib.Path('template.rbx.tex'),
        description='Path to the template that should be used to render the rbx-tex blocks.',
    )


class TexToPDF(BaseModel):
    """Configures the conversion between LaTeX and PDF using pdfLaTeX."""

    type: Literal[ConversionType.TexToPDF]


class JinjaTeX(BaseModel):
    type: Literal[ConversionType.JinjaTeX]


### Joiner types.
class JoinerType(str, Enum):
    TexToPDF = 'tex2pdf'
    """Join contest tex and problem texs to PDF using pdfLaTeX."""

    def __repr__(self):
        return str.__repr__(self.value)


### Joiner nodes.
class JoinTexToPDF(BaseModel):
    """Configures the joining of contest and problem texes to PDF."""

    type: Literal[JoinerType.TexToPDF]


ConversionStep = Union[TexToPDF, JinjaTeX, rbxToTeX, rbxMarkdownToTeX]
Joiner = JoinTexToPDF


### Statement types
class StatementType(AutoEnum):
    rbxTeX = alias('rbx-tex')  # type: ignore
    """Statement written in rbxTeX format."""

    rbxMarkdown = alias('rbxMd', 'rbx-markdown', 'rbx-md')  # type: ignore
    """Statement written in rbxMarkdown format."""

    TeX = alias('tex')
    """Statement written in pure LaTeX format."""

    JinjaTeX = alias('jinja-tex')
    """Statement written in LaTeX format with Jinja2 expressions."""

    PDF = alias('pdf')
    """Statement is a PDF."""

    def get_file_suffix(self) -> str:
        if self == StatementType.TeX:
            return '.tex'
        if self == StatementType.rbxTeX:
            return '.rbx.tex'
        if self == StatementType.rbxMarkdown:
            return '.rbx.md'
        if self == StatementType.JinjaTeX:
            return '.jinja.tex'
        if self == StatementType.PDF:
            return '.pdf'
        raise ValueError(f'Unknown statement type: {self}')


class Statement(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = FNameField(description='Name of this statement.')

    extends: Optional[str] = FNameField(
        default=None,
        description='Name of the statement that this statement extends.',
    )

    language: StatementLanguage = Field(
        default='en', description='Language code of this statement (ISO 639-1).'
    )

    title: str = Field(
        default='', description='Name of the problem, as it appears in the statement.'
    )

    path: pathlib.Path = Field(
        default_factory=pathlib.Path, description='Path to the input statement file.'
    )

    type: StatementType = Field(
        default=StatementType.rbxTeX, description='Type of the input statement file.'
    )

    steps: List[Annotated[ConversionStep, Field(discriminator='type')]] = Field(
        default=[],
        description="""
Describes a sequence of conversion steps that should be applied to the statement file.

Usually, it is not necessary to specify these, as they can be inferred from the
input statement type and the output statement type, but you can use this to force
certain conversion steps to happen.
""",
    )

    configure: List[Annotated[ConversionStep, Field(discriminator='type')]] = Field(
        default=[],
        description="""
Configure how certain conversion steps should happen when applied to the statement file.

Different from the `steps` field, this does not force the steps to happen, but rather only
configure them in case they are applied.
""",
    )

    assets: List[str] = Field(
        default=[],
        description="""
Assets relative to the package directory that should be included while building
the statement. Files will be included in the same folder as the statement file, preserving
their relativeness. Can be glob pattern as well, such as `imgs/*.png`.
""",
    )

    vars: RecVars = Field(
        default={},
        description='Variables to be used in the statement.',
    )

    @property
    def expanded_vars(self) -> Vars:
        return expand_vars(self.vars)
